#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" Provides abstraction for file writer """

import weakref
import time
import pytz
import datetime
import threading
import numpy


#: (:mod:`PNIWriter` or :mod:`H5PYWriter`or :mod:`H5CppWriter`)
#    default writer module
writer = None

#: (:class:`threading.Lock`) writer module
writerlock = threading.Lock()


def open_file(filename, readonly=False, **pars):
    """ open the new file

    :param filename: file name
    :type filename: :obj:`str`
    :param readonly: readonly flag
    :type readonly: :obj:`bool`
    :param pars: parameters
    :type pars: :obj:`dict` < :obj:`str`, :obj:`str`>
    :returns: file object
    :rtype: :class:`FTFile`
    """
    if 'writer' in pars.keys():
        wr = pars.pop('writer')
    else:
        with writerlock:
            wr = writer
    fl = wr.open_file(filename, readonly, **pars)
    if hasattr(fl, "writer"):
        fl.writer = wr
    return fl


def create_file(filename, overwrite=False, **pars):
    """ create a new file

    :param filename: file name
    :type filename: :obj:`str`
    :param overwrite: overwrite flag
    :type overwrite: :obj:`bool`
    :param pars: parameters
    :type pars: :obj:`dict` < :obj:`str`, :obj:`str`>
    :returns: file object
    :rtype: :class:`FTFile`
    """
    if 'writer' in pars.keys():
        wr = pars.pop('writer')
    else:
        with writerlock:
            wr = writer
    fl = wr.create_file(filename, overwrite, **pars)
    if hasattr(fl, "writer"):
        fl.writer = wr
    return fl


def link(target, parent, name):
    """ create link

    :param target: file name
    :type target: :obj:`str`
    :param parent: parent object
    :type parent: :class:`FTObject`
    :param name: link name
    :type name: :obj:`str`
    :returns: link object
    :rtype: :class:`FTLink`
    """
    node = parent
    wr = None
    while node:
        if hasattr(node, "writer"):
            wr = node.writer
            break
        else:
            if hasattr(node, "parent"):
                node = node.parent
            else:
                break
    if not wr:
        with writerlock:
            wr = writer
    return wr.link(target, parent, name)


def get_links(parent):
    """ get links

    :param parent: parent object
    :type parent: :class:`FTObject`
    :returns: list of link objects
    :rtype: :obj: `list` <:class:`FTLink`>
    """
    node = parent
    wr = None
    while node:
        if hasattr(node, "writer"):
            wr = node.writer
            break
        else:
            if hasattr(node, "parent"):
                node = node.parent
            else:
                break
    if not wr:
        with writerlock:
            wr = writer
    return wr.get_links(parent)


def data_filter(parent=None):
    """ create deflate filter

    :param parent: parent object
    :type parent: :class:`FTObject`
    :returns: deflate filter object
    :rtype: :class:`FTDeflate`
    """
    node = parent
    wr = None
    while node:
        if hasattr(node, "writer"):
            wr = node.writer
            break
        else:
            if hasattr(node, "parent"):
                node = node.parent
            else:
                break
    if not wr:
        with writerlock:
            wr = writer
    return wr.data_filter()


deflate_filter = data_filter


def setwriter(wr):
    """ sets writer

    :param wr: writer module
    :type wr: :mod:`PNIWriter` or :mod:`H5PYWriter` or :mod:`H5CppWriter`
    """
    global writer
    with writerlock:
        writer = wr


class FTObject(object):

    """ virtual file tree object
    """

    def __init__(self, h5object, tparent=None):
        """ constructor

        :param h5object: h5 object
        :type h5object: :obj:`any`
        :param tparent: tree parent
        :type tparent: :obj:`FTObject`
        """
        #: (:obj:`any`) h5 object
        self._h5object = h5object
        #: (:obj:`FTObject`) tree parent
        self._tparent = tparent
        #: (:obj:`list` < :obj:`FTObject` > ) weak references of children
        self.__tchildren = []
        if tparent:
            tparent.append(self)

    def append(self, child):
        """ append child weakref

        :param tparent: tree parent
        :type tparent: :obj:`FTObject`
        """
        self.__tchildren.append(weakref.ref(child))

    def __del__(self):
        """ removes weakref in parent object
        """
        if self._tparent:
            self._tparent.reload()

    def reload(self):
        """ reload a list of valid children
        """
        if self.__tchildren:
            self.__tchildren = [
                kd for kd in self.__tchildren if kd() is not None]

    def close(self):
        """ close element
        """
        for ch in self.__tchildren:
            if ch() is not None:
                ch().close()

    def _reopen(self):
        """ reopen elements and children
        """
        self.__tchildren = [ch for ch in self.__tchildren if ch() is not None]
        for ch in self.__tchildren:
            ch().reopen()

    @property
    def parent(self):
        """ return the parent object

        :returns: file tree group
        :rtype: :class:`FTGroup`
        """
        return self._tparent

    @property
    def h5object(self):
        """ provide object of native library

        :returns: pni object
        :rtype: :obj:`any`
        """
        return self._h5object

    @property
    def is_valid(self):
        """ check if attribute is valid

        :returns: valid flag
        :rtype: :obj:`bool`
        """
        return True


def first(array):
    """  get first element if the only

    :param array: numpy array
    :type array: :class:`numpy.ndarray`
    :returns: first element of the array
    :type array: :obj:`any`
    """
    if isinstance(array, numpy.ndarray) and len(array) == 1:
        return array[0]
    else:
        return array


class FTFile(FTObject):

    """ file tree file
    """

    def __init__(self, h5object, filename):
        """ constructor

        :param h5object: pni object
        :type h5object: :obj:`any`
        :param filename:  file name
        :type filename: :obj:`str`
        :param writer: writer module
        :type writer: :mod:`PNIWriter` or :mod:`H5PYWriter`
                        or :mod:`H5CppWriter`
        """
        FTObject.__init__(self, h5object, None)
        #: (:obj:`str`) file name
        self.name = filename
        #: (:mod:`PNIWriter` or :mod:`H5PYWriter` or :mod:`H5CppWriter`)
        # writer module
        self.writer = None

    def root(self):
        """ root object

        :returns: parent object
        :rtype: :class:`FTGroup`
        """

    def flush(self):
        """ flash the data
        """

    @property
    def readonly(self):
        """ check if file is readonly

        :returns: readonly flag
        :rtype: :obj:`bool`
        """

    def hasswmr(self):
        """ if has swmr_mode

        :returns: has swmr_mode
        :rtype: :obj:`bool`
        """
        return False

    def reopen(self, readonly=False, swmr=False, libver=None):
        """ reopen attribute

        :param readonly: readonly flag
        :type readonly: :obj:`bool`
        :param swmr: swmr flag
        :type swmr: :obj:`bool`
        :param libver:  library version, default: 'latest'
        :type libver: :obj:`str`
        """
        FTObject._reopen(self)

    @classmethod
    def currenttime(cls):
        """ returns current time string

        :returns: current time
        :rtype: :obj:`str`
        """
        tzone = time.tzname[0]
        tz = pytz.timezone(tzone)
        fmt = '%Y-%m-%dT%H:%M:%S.%f%z'
        starttime = tz.localize(datetime.datetime.now())
        return str(starttime.strftime(fmt))


class FTGroup(FTObject):

    """ file tree group
    """

    def __init__(self, h5object, tparent=None):
        """ constructor

        :param h5object: pni object
        :type h5object: :obj:`any`
        :param tparent: treee parent
        :type tparent: :obj:`FTObject`
        """
        FTObject.__init__(self, h5object, tparent)
        # : (:obj:`int`) current file id
        self.currentfileid = 0
        #: (:obj:`int`) steps per file
        self.stepsperfile = 0

    def open(self, name):
        """ open a file tree element

        :param name: element name
        :type name: :obj:`str`
        :returns: file tree object
        :rtype: :class:`FTObject`
        """

    def open_link(self, name):
        """ open a file tree element as link

        :param name: element name
        :type name: :obj:`str`
        :returns: file tree object
        :rtype: :class:`FTObject`
        """

    def create_group(self, n, nxclass=""):
        """ open a file tree element

        :param n: group name
        :type n: :obj:`str`
        :param nxclass: group type
        :type nxclass: :obj:`str`
        :returns: file tree group
        :rtype: :class:`FTGroup`
        """

    def create_field(self, name, type_code,
                     shape=None, chunk=None, dfilter=None):
        """ open a file tree element

        :param n: group name
        :type n: :obj:`str`
        :param type_code: nexus field type
        :type type_code: :obj:`str`
        :param shape: shape
        :type shape: :obj:`list` < :obj:`int` >
        :param chunk: chunk
        :type chunk: :obj:`list` < :obj:`int` >
        :param dfilter: filter deflater
        :type dfilter: :class:`FTDeflate`
        :returns: file tree field
        :rtype: :class:`FTField`
        """

    @property
    def size(self):
        """ group size

        :returns: group size
        :rtype: :obj:`int`
        """

    @property
    def attributes(self):
        """ return the attribute manager

        :returns: attribute manager
        :rtype: :class:`FTAttributeManager`
        """

    def exists(self, name):
        """ if child exists

        :param name: child name
        :type name: :obj:`str`
        :returns: existing flag
        :rtype: :obj:`bool`
        """

    def names(self):
        """ read the child names

        :returns: pni object
        :rtype: :obj:`list` <`str`>
        """

    def reopen(self):
        """ reopen attribute
        """
        FTObject._reopen(self)


class FTField(FTObject):

    """ file tree file
    """

    def __init__(self, h5object, tparent=None):
        """ constructor

        :param h5object: pni object
        :type h5object: :obj:`any`
        :param tparent: treee parent
        :type tparent: :obj:`FTObject`
        """
        FTObject.__init__(self, h5object, tparent)

    @property
    def attributes(self):
        """ return the attribute manager

        :returns: attribute manager
        :rtype: :class:`FTAttributeManager`
        """

    def grow(self, dim=0, ext=1):
        """ grow the field

        :param dim: growing dimension
        :type dim: :obj:`int`
        :param dim: size of the grow
        :type dim: :obj:`int`
        """

    def refresh(self):
        """ refresh the field

        :returns: refreshed
        :rtype: :obj:`bool`
        """

    def read(self):
        """ read the field value

        :returns: pni object
        :rtype: :obj:`any`
        """

    def write(self, o):
        """ write the field value

        :param o: pni object
        :type o: :obj:`any`
        """

    def __setitem__(self, t, o):
        """ set value

        :param t: slice tuple
        :type t: :obj:`tuple`
        :param o: pni object
        :type o: :obj:`any`
        """

    def __getitem__(self, t):
        """ get value

        :param t: slice tuple
        :type t: :obj:`tuple`
        :returns: pni object
        :rtype: :obj:`any`
        """

    @property
    def dtype(self):
        """ field data type

        :returns: field data type
        :rtype: :obj:`str`
        """

    @property
    def shape(self):
        """ field shape

        :returns: field shape
        :rtype: :obj:`list` < :obj:`int` >
        """

    @property
    def size(self):
        """ field size

        :returns: field size
        :rtype: :obj:`int`
        """

    def reopen(self):
        """ reopen attribute
        """
        FTObject._reopen(self)


class FTLink(FTObject):

    """ file tree link
    """

    def __init__(self, h5object, tparent=None):
        """ constructor

        :param h5object: pni object
        :type h5object: :obj:`any`
        :param tparent: treee parent
        :type tparent: :obj:`FTObject`
        """
        FTObject.__init__(self, h5object, tparent)

    @property
    def target_path(self):
        """ target path

        :returns: target path
        :rtype: :obj:`str`
        """

    def refresh(self):
        """ refresh the link

        :returns: refreshed
        :rtype: :obj:`bool`
        """

    def reopen(self):
        """ reopen attribute
        """
        FTObject._reopen(self)


class FTDataFilter(FTObject):

    """ file tree data filter
    """

    def __init__(self, h5object=None, tparent=None):
        """ constructor

        :param h5object: pni object
        :type h5object: :obj:`any`
        :param tparent: treee parent
        :type tparent: :obj:`FTObject`
        """
        FTObject.__init__(self, h5object, tparent)
        #: (:obj:`bool`) compression shuffle
        self._shuffle = False
        #: (:obj:`int`) compression rate
        self._rate = 0
        #: (:obj:`int`) compression filter id
        self._filterid = 1
        #: (:obj:`tuple` <:obj:`int`>) compression options
        self._options = tuple()

    @property
    def options(self):
        """ getter for compression options

        :returns: compression options
        :rtype: :obj:`tuple` <:obj:`int`>
        """
        return self._options

    @options.setter
    def options(self, value):
        """ setter for compression options

        :param value: compression options
        :type value: :obj:`tuple` <:obj:`int`>
        """
        self._options = value

    @property
    def rate(self):
        """ getter for compression rate

        :returns: compression rate
        :rtype: :obj:`int`
        """
        return self._rate

    @rate.setter
    def rate(self, value):
        """ setter for compression rate

        :param value: compression rate
        :type value: :obj:`int`
        """
        self._rate = value

    @property
    def filterid(self):
        """ getter for compression filter id

        :returns: compression rate
        :rtype: :obj:`int`
        """
        return self._filterid

    @filterid.setter
    def filterid(self, value):
        """ setter for compression filter id

        :param value: compression filter id
        :type value: :obj:`int`
        """
        self._filterid = value

    @property
    def shuffle(self):
        """ getter for compression shuffle

        :returns: compression shuffle
        :rtype: :obj:`bool`
        """
        return self._shuffle

    @shuffle.setter
    def shuffle(self, value):
        """ setter for compression shuffle

        :param value: compression shuffle
        :type value: :obj:`bool`
        """
        self._shuffle = value

    def reopen(self):
        """ reopen attribute
        """
        FTObject._reopen(self)


class FTDeflate(FTDataFilter):
    pass


class FTAttributeManager(FTObject):

    """ file tree attribute
    """

    def __init__(self, h5object, tparent=None):
        """ constructor

        :param h5object: pni object
        :type h5object: :obj:`any`
        :param tparent: treee parent
        :type tparent: :obj:`FTObject`
        """
        FTObject.__init__(self, h5object, tparent)

    def create(self, name, dtype, shape=None, overwrite=False):
        """ create a new attribute

        :param name: attribute name
        :type name: :obj:`str`
        :param dtype: attribute type
        :type dtype: :obj:`str`
        :param shape: attribute shape
        :type shape: :obj:`list`< :obj:`int`>
        :param overwrite: overwrite flag
        :type overwrite: :obj:`bool`
        :returns: attribute object
        :rtype: :class:`FTAtribute`
        """

    def __len__(self):
        """ number of attributes

        :returns: number of attributes
        :rtype: :obj:`int`
        """

    def __getitem__(self, name):
        """ get value

        :param name: attribute name
        :type name: :obj:`str`
        :returns: attribute object
        :rtype: :class:`FTAtribute`
        """

    def reopen(self):
        """ reopen attribute
        """
        FTObject._reopen(self)


class FTAttribute(FTObject):

    """ virtual file tree attribute
    """

    def __init__(self, h5object, tparent=None):
        """ constructor

        :param h5object: pni object
        :type h5object: :obj:`any`
        :param tparent: treee parent
        :type tparent: :obj:`FTObject`
        """
        FTObject.__init__(self, h5object, tparent)

    def read(self):
        """ read attribute value

        :returns: python object
        :rtype: :obj:`any`
        """

    def write(self, o):
        """ write attribute value

        :param o: python object
        :type o: :obj:`any`
        """

    def __setitem__(self, t, o):
        """ write attribute value

        :param t: slice tuple
        :type t: :obj:`tuple`
        :param o: python object
        :type o: :obj:`any`
        """

    def __getitem__(self, t):
        """ read attribute value

        :param t: slice tuple
        :type t: :obj:`tuple`
        :returns: python object
        :rtype: :obj:`any`
        """

    @property
    def dtype(self):
        """ attribute data type

        :returns: attribute data type
        :rtype: :obj:`str`
        """

    @property
    def shape(self):
        """ attribute shape

        :returns: attribute shape
        :rtype: :obj:`list` < :obj:`int` >
        """

    def reopen(self):
        """ reopen attribute
        """
        FTObject._reopen(self)



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

"""Command-line tool for creating NXSConfigServer configuration of Nexus Files
"""

import os
import sys
import argparse


from nxstools.nxsargparser import (Runner, NXSArgParser, ErrorException)
from nxstools.nxsdevicetools import (
    checkServer, getAttributes, getServerTangoHost)
from nxstools.nxscreator import (
    TangoDSCreator, ClientDSCreator, WrongParameterError,
    DeviceDSCreator, OnlineDSCreator, OnlineCPCreator, CPExistsException,
    DSExistsException,
    StandardCPCreator, ComponentCreator, CompareOnlineDS, PoolDSCreator)


#: (:obj:`bool`) True if PyTango available
PYTANGO = False
try:
    __import__("PyTango")
    PYTANGO = True
except Exception:
    pass


class TangoDS(Runner):

    """ tangods runner"""

    #: (:obj:`str`) command description
    description = "create tango datasources"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " * with -b: datasources are created" \
        + " in Configuration Server database\n" \
        + " * without -b: datasources are created" \
        + " on the local filesystem in -d <directory> \n" \
        + " * default: <directory> is '.' \n" \
        + "            <server> is taken from Tango DB\n" \
        + "            <datasource> is 'exp_mot' \n" \
        + "            <host>, <port> are taken from <server>\n" \
        + " examples:\n" \
        + "       nxscreate tangods -f1 -l2 -v p09/motor/exp. -s exp_mot \n" \
        + "       nxscreate tangods -f1 -l32 -v p02/motor/eh1a. -s exp_mot" \
        + "-b \n" \
        + "       nxscreate tangods -f1 -l32 -v p01/motor/oh1. " \
        + "-s exp_mot -b\n" \
        + "       nxscreate tangods -v petra/globals/keyword " \
        + "-s source_current -u haso228 -t 10000 \\ \n "\
        + "                        -a BeamCurrent " \
        + "-b -r p09/nxsconfigserver/haso228 -o -g __CLIENT__\n" \
        + "       nxscreate tangods -f1 -l8  -v pXX/slt/exp. -s slt_exp_ -u" \
        + " hasppXX.desy.de -b \n" \
        + "\n"

    def create(self):
        """ creates parser

        """
        parser = self._parser
        parser.add_argument(
            "-v", "--device-prefix",
            help="device prefix, i.e. exp_c (mandatory)",
            dest="device", default="")
        parser.add_argument(
            "-f", "--first", help="first index",
            dest="first", default="1")
        parser.add_argument(
            "-l", "--last",
            help="last index",
            dest="last", default=None)

        parser.add_argument(
            "-a", "--attribute",
            help="tango attribute name",
            dest="attribute", default="Position")

        parser.add_argument(
            "-s", "--datasource-prefix",
            help="datasource-prefix "
            "(useful for avoiding duplicated datasource names)",
            dest="datasource", default="exp_mot")

        parser.add_argument(
            "-o", "--overwrite", action="store_true",
            default=False, dest="overwrite",
            help="overwrite existing datasources")
        parser.add_argument(
            "-d", "--directory",
            help="output datasource directory",
            dest="directory", default=".")
        parser.add_argument(
            "-x", "--file-prefix",
            help="file prefix, i.e. counter",
            dest="file", default="")
        parser.add_argument(
            "-u", "--host",
            help="tango host name",
            dest="host", default="")
        parser.add_argument(
            "-t", "--port",
            help="tango host port",
            dest="port", default="10000")

        parser.add_argument(
            "-b", "--database", action="store_true",
            default=False, dest="database",
            help="store datasources in "
            "Configuration Server database")
        parser.add_argument(
            "-g", "--group",
            help="device group name",
            dest="group", default="")
        parser.add_argument(
            "-e", "--elementtype",
            help="element type, i.e. attribute, property or command",
            dest="elementtype", default="attribute")

        parser.add_argument(
            "-r", "--server", dest="server",
            help="configuration server device name")
        return parser

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """

        parser = self._parser
        if (options.database and not options.server) or \
           (not options.host and not options.server):
            if not PYTANGO:
                sys.stderr.write("Info: No PyTango installed\n")
                sys.stderr.flush()
                sys.exit(255)
            options.server = checkServer()
            if options.database and not options.server:
                parser.print_help()
                sys.exit(0)

        if not options.host:
            if not PYTANGO:
                sys.stderr.write(
                    "Info: No Tango Host or PyTango installed\n")
                sys.stderr.flush()
                sys.exit(255)
            hostport = getServerTangoHost(options.server)
            options.host, options.port = hostport.split(":")

        if options.database:
            print("CONFIG SERVER: %s" % str(options.server))
        else:
            print("OUTPUT DIRECTORY: %s" % str(options.directory))

        creator = TangoDSCreator(options, [])
        try:
            creator.create()
        except WrongParameterError as e:
            sys.stderr.write(str(e))
            sys.stderr.flush()
            parser.print_help()
            sys.exit(255)


class DeviceDS(Runner):

    """ deviceds runner"""
    #: (:obj:`str`) command description
    description = "create datasources for all device attributes"
    #: (:obj:`str`) command epilog
    epilog = "" \
             + " * without <dv_attr1>: datasources for all" \
             + " attributes are created\n" \
             + " * with -b: datasources are created" \
             + " in Configuration Server database\n" \
             + " * without -b: datasources are created" \
             + " on the local filesystem in -d <directory> \n" \
             + " * default: <directory> is '.' \n" \
             + "            <server> is taken from Tango DB\n" \
             + "            <datasource> is 'exp_mot' \n" \
             + "            <host>, <port> are taken from <server>\n" \
             + " examples:\n" \
             + "       nxscreate deviceds  -v p09/pilatus/haso228k \n" \
             + "       nxscreate deviceds  -v p09/lambda2m/haso228k -u" \
             + "haslambda -b \n" \
             + "       nxscreate deviceds  -v p09/pilatus300k/haso228k -b" \
             + " -s pilatus300k_ RoI Energy ExposureTime\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument("-v", "--device",
                            help="device, i.e. p09/pilatus300k/01 (mandatory)",
                            dest="device", default="")
        parser.add_argument("-s", "--datasource-prefix",
                            help="datasource-prefix"
                            " (useful for avoiding duplicated"
                            " datasource names)",
                            dest="datasource", default="")

        parser.add_argument("-o", "--overwrite", action="store_true",
                            default=False, dest="overwrite",
                            help="overwrite existing datasources")
        parser.add_argument("-d", "--directory",
                            help="output datasource directory",
                            dest="directory", default=".")
        parser.add_argument("-x", "--file-prefix",
                            help="file prefix, i.e. counter",
                            dest="file", default="")
        parser.add_argument("-u", "--host",
                            help="tango host name",
                            dest="host", default="")
        parser.add_argument("-t", "--port",
                            help="tango host port",
                            dest="port", default="10000")

        parser.add_argument("-b", "--database", action="store_true",
                            default=False, dest="database",
                            help="store components in Configuration"
                            " Server database")

        parser.add_argument("-n", "--no-group", action="store_true",
                            default=False, dest="nogroup",
                            help="don't create common group with a name of"
                            " datasource prefix")

        parser.add_argument("-r", "--server", dest="server",
                            help="configuration server device name")

        parser.add_argument('args', metavar='attribute_name',
                            type=str, nargs='*',
                            help='attribute names of the tango device')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """

        parser = self._parser
        if (options.database and not options.server) or \
           (not options.host and not options.server):
            if not PYTANGO:
                sys.stderr.write("Info: No PyTango installed\n")
                sys.stderr.flush()
                sys.exit(255)

            options.server = checkServer()
            if options.database and not options.server:
                parser.print_help()
                sys.exit(0)

        if not options.host:
            if not PYTANGO:
                sys.stderr.write("Info: No Tango Host or PyTango installed\n")
                sys.stderr.flush()
                sys.exit(255)
            hostport = getServerTangoHost(options.server)
            options.host, options.port = hostport.split(":")

        if options.database:
            print("CONFIG SERVER: %s" % options.server)
        else:
            print("OUTPUT DIRECTORY: %s" % options.directory)

        if not options.device.strip():
            parser.print_help()
            sys.exit(255)

        if options.args:
            aargs = list(options.args)
        else:
            if not PYTANGO:
                sys.stderr.write("CollCompCreator No PyTango installed\n")
                sys.stderr.flush()
                parser.print_help()
                sys.exit(255)
            aargs = getAttributes(options.device, options.host, options.port)

        creator = DeviceDSCreator(options, aargs)
        creator.create()


class OnlineDS(Runner):

    """ onlineds runner"""

    #: (:obj:`str`) command description
    description = "create datasources from online.xml file"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " * with -b: datasources are created" \
        + " in Configuration Server database\n" \
        + " * with -d <directory>: datasources are created" \
        + " on the local filesystem\n" \
        + " * without -b or -d <directory>: run in the test mode\n" \
        + " * default: <inputFile> is '/online_dir/online.xml' \n" \
        + "            <server> is taken from Tango DB\n\n" \
        + " `onlineds` overwrites existing datasources\n\n" \
        + " examples:\n" \
        + "       nxscreate onlineds -b  \n" \
        + "       nxscreate onlineds -b -t \n" \
        + "       nxscreate onlineds -d /home/user/xmldir \n" \
        + "       nxscreate onlineds \n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument("-b", "--database", action="store_true",
                            default=False, dest="database",
                            help="store components in"
                            " Configuration Server database")
        parser.add_argument("-t", "--noclientlike", action="store_false",
                            default=True, dest="clientlike",
                            help="set motor tango datasources to "
                            "be no __CLIENT__ like, i.e. pure tango-like")
        parser.add_argument("-d", "--directory",
                            help="output directory where"
                            " datasources will be saved",
                            dest="directory", default="")
        parser.add_argument("-n", "--nolower", action="store_false",
                            default=True, dest="lower",
                            help="do not change aliases into lower case")
        parser.add_argument("-r", "--server", dest="server",
                            help="configuration server device name")
        parser.add_argument("-x", "--file-prefix",
                            help="file prefix, i.e. counter",
                            dest="file", default="")
        parser.add_argument("-e", "--external",
                            help="external configuration server",
                            dest="external", default="")
        parser.add_argument("-p", "--xml-package", dest="xmlpackage",
                            help="xml template package")
        parser.add_argument("-c", "--clientlike", action="store_true",
                            default=False, dest="oldclientlike",
                            help="set motor tango datasources to "
                            "be __CLIENT__ like (deprecated)")
        parser.add_argument("--verbose", action="store_true",
                            default=False, dest="verbose",
                            help="printout verbose mode")

    def postauto(self):
        """ creates parser
        """
        self._parser.add_argument('args', metavar='online_file',
                                  type=str, nargs='?',
                                  help='online.xml file')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """

        parser = self._parser
        if not PYTANGO:
            sys.stderr.write("Info: No PyTango installed\n")
            sys.stderr.flush()
            sys.exit(255)

        if not options.server:
            options.server = checkServer()
            if not options.server and options.database:
                parser.print_help()
                sys.exit(0)
        args = [options.args] if options.args else []
        if not len(args) and os.path.isfile('/online_dir/online.xml'):
            args = ['/online_dir/online.xml']

        if not len(args):
            parser.print_help()
            sys.exit(255)

        print("INPUT: %s" % args)
        if options.directory:
            print("OUTPUT DIR: %s" % options.directory)
        elif options.database:
            print("SERVER: %s" % options.server)
        else:
            print("TEST MODE: %s" % options.server)

        creator = OnlineDSCreator(options, args)
        creator.create()


class PoolDS(Runner):

    """ poolds runner"""

    #: (:obj:`str`) command description
    description = "create datasources from sardana pool device"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " * with -b: datasources are created" \
        + " in Configuration Server database\n" \
        + " * with -d <directory>: datasources are created" \
        + " on the local filesystem\n" \
        + " * without -b or -d <directory>: run in the test mode\n" \
        + " * default: <channel> is 'ALL' \n" \
        + "            <server> is taken from Tango DB\n\n" \
        + "            <pool> is taken from Tango DB\n\n" \
        + " `poolds` overwrites existing datasources\n\n" \
        + " examples:\n" \
        + "       nxscreate poolds -b  \n" \
        + "       nxscreate poolds -b -t \n" \
        + "       nxscreate poolds -d -p p09/pool/haso228 \n" \
        + "       nxscreate poolds -b Motor CTExpChannel \n" \
        + "       nxscreate poolds -b mot01 mot03 \n" \
        + "       nxscreate poolds \n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument("-b", "--database", action="store_true",
                            default=False, dest="database",
                            help="store components in"
                            " Configuration Server database")
        parser.add_argument("-t", "--noclientlike", action="store_false",
                            default=True, dest="clientlike",
                            help="set motor tango datasources to "
                            "be no __CLIENT__ like, i.e. pure tango-like")
        parser.add_argument("-d", "--directory",
                            help="output directory where"
                            " datasources will be saved",
                            dest="directory", default="")
        parser.add_argument("-n", "--nolower", action="store_false",
                            default=True, dest="lower",
                            help="do not change aliases into lower case")
        parser.add_argument("-r", "--server", dest="server",
                            help="configuration server device name")
        parser.add_argument("-x", "--file-prefix",
                            help="file prefix, i.e. counter",
                            dest="file", default="")
        parser.add_argument("-p", "--pool",
                            help="sardana pool device name",
                            dest="pool", default="")

    def postauto(self):
        """ creates parser
        """
        self._parser.add_argument('args', metavar='channel',
                                  type=str, nargs='*',
                                  help='sardana pool channel types '
                                  'or name of aliases, '
                                  'e.g. Motor CTExpChannel ZeroDExpChannel '
                                  'OneDExpChannel TwoDExpChannel '
                                  ' PseudoMotor PseudoCounter ALL')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """

        parser = self._parser
        if not PYTANGO:
            sys.stderr.write("Info: No PyTango installed\n")
            sys.stderr.flush()
            sys.exit(255)
        if not options.server:
            options.server = checkServer()
            if options.database and not options.server:
                parser.print_help()
                sys.exit(0)

        if not options.pool:
            options.pool = checkServer("Pool")
            if not options.pool:
                parser.print_help()
                sys.exit(0)
        args = options.args

        print("INPUT: %s" % args)
        if options.directory:
            print("OUTPUT DIR: %s" % options.directory)
        elif options.database:
            print("SERVER: %s" % options.server)
        else:
            print("TEST MODE: %s" % options.server)

        creator = PoolDSCreator(options, args)
        creator.create()


class Compare(Runner):

    """ compare runner"""

    #: (:obj:`str`) command description
    description = "compare two online.xml files"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " * default: second file <online_file> is " \
        + "'/online_dir/online.xml' \n" \
        + "            if only file is given\n\n" \
        + " examples:\n" \
        + "       nxscreate onlineds /online_dir/online.xml online.xml \n" \
        + "       nxscreate compare online.xml \n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument("-n", "--nolower", action="store_false",
                            default=True, dest="lower",
                            help="do not change aliases into lower case")

    def postauto(self):
        """ creates parser
        """
        self._parser.add_argument('args', metavar='online_file',
                                  type=str, nargs='+',
                                  help='online.xml files')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        args = options.args if options.args else []
        parser = self._parser

        if len(args) == 1 and os.path.isfile('/online_dir/online.xml'):
            args.append('/online_dir/online.xml')

        if len(args) == 1:
            parser.print_help()
            sys.exit(255)

        creator = CompareOnlineDS(options, args)
        creator.compare()


class OnlineCP(Runner):

    """ onlinecp runner"""
    #: (:obj:`str`) command description
    description = "create component from online.xml file"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " * without '-c <component>': show a list of possible components\n" \
        + " * with -b: datasources are created" \
        + " in Configuration Server database\n" \
        + " * without -b: datasources are created" \
        + " on the local filesystem in -d <directory> \n" \
        + " * default: <directory> is '.' \n" \
        + " * default: <inputFile> is '/online_dir/online.xml' \n" \
        + "            <server> is taken from Tango DB\n" \
        + " examples:\n" \
        + "       nxscreate onlinecp  \n" \
        + "       nxscreate onlinecp -c pilatus -b \n" \
        + "       nxscreate onlinecp -c lambda -d /home/user/xmldir/ \n\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument("-c", "--component",
                            help="component name" +
                            "related to the device name from <inputFile>",
                            dest="component", default="")
        parser.add_argument("-b", "--database", action="store_true",
                            default=False, dest="database",
                            help="store components in"
                            "Configuration Server database")
        parser.add_argument("-r", "--server", dest="server",
                            help="configuration server device name")
        parser.add_argument("-n", "--nolower", action="store_false",
                            default=True, dest="lower",
                            help="do not change aliases into lower case")
        parser.add_argument("-o", "--overwrite", action="store_true",
                            default=False, dest="overwrite",
                            help="overwrite existing component")
        parser.add_argument("-d", "--directory",
                            help="output datasource directory",
                            dest="directory", default=".")
        parser.add_argument("-x", "--file-prefix",
                            help="file prefix, i.e. counter",
                            dest="file", default="")
        parser.add_argument("-e", "--external",
                            help="external configuration server",
                            dest="external", default="")
        parser.add_argument("-p", "--xml-package", dest="xmlpackage",
                            help="xml template package")
        parser.add_argument("-y", "--entryname", dest="entryname",
                            help="entry group name (prefix)", default="scan")
        parser.add_argument("-i", "--insname", dest="insname",
                            help="instrument group name", default="instrument")

    def postauto(self):
        """ creates parser
        """
        self._parser.add_argument('args', metavar='online_file',
                                  type=str, nargs='?',
                                  help='online.xml file')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """

        parser = self._parser
        if not PYTANGO:
            sys.stderr.write("Info: No PyTango installed\n")
            sys.stderr.flush()
            sys.exit(255)

        if not options.server:
            options.server = checkServer()
            if not options.server and options.database:
                parser.print_help()
                sys.exit(0)

        args = [options.args] if options.args else []
        if not len(args) and os.path.isfile('/online_dir/online.xml'):
            args = ['/online_dir/online.xml']

        if not len(args):
            parser.print_help()
            sys.exit(255)

        print("INPUT: %s" % args[0])
        if options.database:
            print("SERVER: %s" % options.server)
        else:
            print("OUTPUT DIR: %s" % options.directory)

        creator = OnlineCPCreator(options, args)
        if options.component:
            try:
                creator.create()
            except CPExistsException as e:
                print(str(e))
                sys.exit(255)
        else:
            lst = creator.listcomponents()
            print("\nPOSSIBLE COMPONENTS: \n   %s" % " ".join(list(lst)))


class StdComp(Runner):

    """  stdcomp runner"""

    #: (:obj:`str`) command description
    description = "create component from the standard templates"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " * without '-t <type>': show a list of possible" \
        + " component types\n" \
        + " * with '-t <type>  and without -c <component>:" \
        + " show a list of component variables " \
        + "for the given component type\n" \
        + " * with -b: datasources are created" \
        + " in Configuration Server database\n" \
        + " * without -b: datasources are created" \
        + " on the local filesystem in -d <directory> \n" \
        + " * default: <directory> is '.' \n" \
        + " * [name1 value1 [name2 value2] ...] sequence" \
        + "  defines component variable values \n" \
        + " examples:\n" \
        + "       nxscreate stdcomp  \n" \
        + "       nxscreate stdcomp -t source \n" \
        + "       nxscreate stdcomp -t default -c default -m -b\n" \
        + "       nxscreate stdcomp -t slit -c front_slit1 -b" \
        + " xgap slt1x ygap slt1y\n\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument("-c", "--component",
                            help="component name",
                            dest="component", default="")
        parser.add_argument("-t", "--type",
                            help="component type",
                            dest="cptype", default="")
        parser.add_argument("-r", "--server", dest="server",
                            help="configuration server device name")
        parser.add_argument("-p", "--xml-package", dest="xmlpackage",
                            help="xml template package")
        parser.add_argument("-n", "--nolower", action="store_false",
                            default=True, dest="lower",
                            help="do not change aliases into lower case")
        parser.add_argument("-o", "--overwrite", action="store_true",
                            default=False, dest="overwrite",
                            help="overwrite existing component")
        parser.add_argument("-b", "--database", action="store_true",
                            default=False, dest="database",
                            help="store components in"
                            "Configuration Server database")
        parser.add_argument("-d", "--directory",
                            help="output datasource directory",
                            dest="directory", default=".")
        parser.add_argument("-m", "--mandatory", action="store_true",
                            default=False, dest="mandatory",
                            help="set the component as mandatory")
        parser.add_argument("-e", "--external",
                            help="external configuration server",
                            dest="external", default="")
        parser.add_argument("-x", "--file-prefix",
                            help="file prefix, i.e. counter",
                            dest="file", default="")
        parser.add_argument("-y", "--entryname", dest="entryname",
                            help="entry group name (prefix)", default="scan")
        parser.add_argument("-i", "--insname", dest="insname",
                            help="instrument group name", default="instrument")
        parser.add_argument('args', metavar='key value',
                            type=str, nargs='*',
                            help='pairs of (key value) for template variables')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        parser = self._parser
        if not PYTANGO:
            sys.stderr.write("Info: No PyTango installed\n")
            parser.print_help()
            sys.exit(255)

        args = options.args or []

        creator = StandardCPCreator(options, args)
        if options.component and options.cptype:
            if not options.server:
                options.server = checkServer()
                if not options.server and options.database:
                    parser.print_help()
                    sys.exit(0)
            if options.database:
                print("SERVER: %s" % options.server)
            else:
                print("OUTPUT DIR: %s" % options.directory)

            try:
                creator.create()
            except CPExistsException as e:
                print(str(e))
                sys.exit(255)
            except DSExistsException as e:
                print(str(e))
                sys.exit(255)
        elif options.cptype:
            dct = creator.listcomponentvariables()
            print("\nCOMPONENT VARIABLES:")
            for var in sorted(dct.keys()):
                desc = dct[var]
                if not var.startswith('__') and not var.endswith('__'):
                    print("  %s - %s [default: '%s']"
                          % (var, desc['doc'], desc['default']))
        else:
            parser.print_help()
            print("")
            lst = sorted(creator.listcomponenttypes())
            print("\nPOSSIBLE COMPONENT TYPES: \n   %s" % " ".join(lst))


class Comp(Runner):

    """ comp runner"""
    #: (:obj:`str`) command description
    description = "create simple components"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " * with -b: components are created (without datasources)" \
        + " in Configuration Server database\n" \
        + " * without -b: components are created (without datasources)" \
        + " on the local filesystem in -d <directory> \n" \
        + " * default: <directory> is '.' \n" \
        + "            <server> is taken from Tango DB\n" \
        + "            <strategy> is step\n" \
        + "            <type> is NX_FLOAT\n" \
        + "            <chunk> is SCALAR\n" \
        + "            <nexuspath> is " \
        + "\"/\\$var.entryname#'scan'\\$var.serialno:NXentry" \
        + "/instrument/collection/\"\n" \
        + " examples:\n" \
        + "       nxscreate comp counter \n" \
        + "       nxscreate comp -f1 -l -v exp_c01 -b \n" \
        + "       nxscreate comp lambda -d /home/user/xmldir/ \n" \
        + "       nxscreate comp -n " \
        + "\"/\\$var.entryname#'scan'\\$var.serialno:NXentry/instrument/" \
        + "sis3302:NXdetector/collection:NXcollection/\" " \
        + "-v sis3302_1_roi -f1 -l4 " \
        + " -g STEP -t NX_FLOAT64 -k -b -m \n" \
        + "       nxscreate comp -n " \
        + "\"/\\$var.entryname#'scan'\\$var.serialno:NXentry/instrument/" \
        + "eh1_mca01:NXdetector/data\" eh1_mca01 -g STEP -t NX_FLOAT64" \
        + " -i -b -c SPECTRUM\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-v", "--device-prefix", help="device prefix, i.e. exp_c",
            dest="device", default="")
        parser.add_argument(
            "-f", "--first", help="first index",
            dest="first", default="1")
        parser.add_argument(
            "-l", "--last", help="last index",
            dest="last", default=None)
        parser.add_argument(
            "-o", "--overwrite", action="store_true",
            default=False, dest="overwrite",
            help="overwrite existing components")
        parser.add_argument(
            "-a", "--can-fail", action="store_true",
            default=False, dest="canfail",
            help="can fail strategy flag")
        parser.add_argument(
            "-d", "--directory",
            help="output component directory",
            dest="directory", default=".")
        parser.add_argument(
            "-x", "--file-prefix",
            help="file prefix, i.e. counter",
            dest="file", default="")

        parser.add_argument(
            "-n", "--nexuspath",
            help="nexus path with field name",
            dest="nexuspath", default="")

        parser.add_argument(
            "-g", "--strategy",
            help="writing strategy, i.e. "
            "STEP, INIT, FINAL, POSTRUN",
            dest="strategy", default="STEP")
        parser.add_argument(
            "-s", "--datasource-prefix",
            help="datasource-prefix or datasourcename",
            dest="datasource", default="")
        parser.add_argument(
            "-t", "--type",
            help="nexus type of the field",
            dest="type", default="NX_FLOAT")
        parser.add_argument(
            "-u", "--units",
            help="nexus units of the field",
            dest="units", default="")

        parser.add_argument(
            "-k", "--links", action="store_true",
            default=False, dest="fieldlinks",
            help="create links with field name")

        parser.add_argument(
            "-i", "--source-links", action="store_true",
            default=False, dest="sourcelinks",
            help="create links with datasource name")

        parser.add_argument(
            "-b", "--database", action="store_true",
            default=False, dest="database",
            help="store components in "
            "Configuration Server database")

        parser.add_argument(
            "-r", "--server", dest="server",
            help="configuration server device name")

        parser.add_argument(
            "-c", "--chunk", dest="chunk",
            default="SCALAR", help="chunk format, "
            "i.e. SCALAR, SPECTRUM, IMAGE")

        parser.add_argument(
            "-m", "--minimal-device", action="store_true",
            default=False, dest="minimal",
            help="device name without first '0'")

        parser.add_argument(
            'args', metavar='component_name',
            type=str, nargs='*',
            help='component names to be created')

        return parser

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        args = options.args or []
        parser = self._parser
        if options.database and not options.server:
            if not PYTANGO:
                sys.stderr.write("Info: No PyTango installed\n")
                sys.stderr.flush()
                sys.exit(255)

            options.server = checkServer()
            if not options.server:
                parser.print_help()
                sys.exit(0)

        if options.database:
            print("CONFIG SERVER: %s" % options.server)
        else:
            print("OUTPUT DIRECTORY: %s" % options.directory)

        creator = ComponentCreator(options, args)
        try:
            creator.create()
        except WrongParameterError as e:
            sys.stderr.write(str(e))
            sys.stderr.flush()
            parser.print_help()
            sys.exit(255)


class ClientDS(Runner):

    """ clientds runner"""
    #: (:obj:`str`) command description
    description = "create client datasources"

    #: (:obj:`str`) command epilog
    epilog = "" \
             + "\n" \
             + " * with -b: datasources are created" \
             + " in Configuration Server database\n" \
             + " * without -b: datasources are created" \
             + " on the local filesystem in -d <directory> \n" \
             + " * default: <directory> is '.' \n" \
             + "            <server> is taken from Tango DB\n" \
             + " examples:\n" \
             + "       nxscreate clientds starttime -b  \n" \
             + "       nxscreate clientds title -d /home/user/xmldir \n" \
             + "       nxscreate clientds -v exp_c -f1 -l4 -b  \n" \
             + "       nxscreate clientds -v " \
             + "hasppXX:10000/expchan/vfcadc_exp/" \
             + " -f1 -l8  -m -b -s exp_vfc\n"

    def create(self):
        """ parser creator
        """
        parser = self._parser
        parser.add_argument("-v", "--device-prefix",
                            help="device prefix, i.e. exp_c "
                            "(mandatory w/o <name1>)",
                            dest="device", default="")
        parser.add_argument("-f", "--first",
                            help="first index (mandatory w/o <name1>)",
                            dest="first", default="1")
        parser.add_argument("-l", "--last",
                            help="last index (mandatory w/o <name1>)",
                            dest="last", default=None)

        parser.add_argument("-o", "--overwrite", action="store_true",
                            default=False, dest="overwrite",
                            help="overwrite existing datasources")
        parser.add_argument("-d", "--directory",
                            help="output datasource directory",
                            dest="directory", default=".")
        parser.add_argument("-x", "--file-prefix",
                            help="file prefix, i.e. counter",
                            dest="file", default="")

        parser.add_argument("-s", "--datasource-prefix",
                            help="datasource prefix, i.e. counter"
                            " (useful for avoiding duplicated "
                            "datasource names)",
                            dest="dsource", default="")

        parser.add_argument("-b", "--database", action="store_true",
                            default=False, dest="database",
                            help="store datasources in "
                            "Configuration Server database")
        parser.add_argument("-m", "--minimal-device", action="store_true",
                            default=False, dest="minimal",
                            help="device name without first '0'")

        parser.add_argument("-r", "--server", dest="server",
                            help="configuration server device name")

        parser.add_argument('args', metavar='device_name',
                            type=str, nargs='*',
                            help='device names to create')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """

        parser = self._parser
        if options.database and not options.server:
            if not PYTANGO:
                sys.stderr.write("Info: No PyTango installed\n")
                sys.stderr.flush()
                sys.exit(255)

            options.server = checkServer()
            if not options.server:
                parser.print_help()
                sys.exit(0)

        if options.database:
            print("CONFIG SERVER: %s" % options.server)
        else:
            print("OUTPUT DIRECTORY: %s" % options.directory)

        creator = ClientDSCreator(options, options.args or [])
        try:
            creator.create()
        except WrongParameterError as e:
            sys.stderr.write(str(e))
            sys.stderr.flush()
            parser.print_help()
            sys.exit(255)


def main():
    """ the main program function
    """
    description = " Command-line tool for creating NXSConfigServer" \
                  + " configuration of Nexus Files"

    epilog = 'For more help:\n  nxscreate <sub-command> -h'
    parser = NXSArgParser(
        description=description, epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.cmdrunners = [('clientds', ClientDS),
                         ('tangods', TangoDS),
                         ('deviceds', DeviceDS),
                         ('onlinecp', OnlineCP),
                         ('onlineds', OnlineDS),
                         ('poolds', PoolDS),
                         ('stdcomp', StdComp),
                         ('comp', Comp),
                         ('compare', Compare)]
    runners = parser.createSubParsers()

    try:
        options = parser.parse_args()
    except ErrorException as e:
        sys.stderr.write("Error: %s\n" % str(e))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    if options.subparser is None:
        sys.stderr.write(
            "Error: %s\n" % str("too few arguments"))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    runners[options.subparser].run(options)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" Command-line tool to ascess to Tango Data Server"""

import sys

import argparse

from .nxsdevicetools import (checkServer, listServers, openServer)
from .nxsargparser import (Runner, NXSArgParser, ErrorException)


class NexusServer(object):

    """ configuration server adapter
    """

    def __init__(self, device):
        """ constructor

        :param device: device name of configuration server
        :type device: :obj:`str`
        """
        #: (:class:`PyTango.DeviceProxy`) NeXus writer device proxy
        self.tdwServer = openServer(device)

    def openFile(self, filename):
        """ opens the h5 file

        :param filename: h5 file name
        :type filename: :obj:`str`
        """
        self.tdwServer.Init()
        self.tdwServer.FileName = str(filename)
        self.tdwServer.OpenFile()

    def setData(self, jsondata):
        """ sets the global JSON data

        :param jsondata: global JSON data
        :type jsondata: :obj:`str`
        """
        self.tdwServer.JSONRecord = str(jsondata)

    def openEntry(self, xmlconfig):
        """ opens an entry

        :param xmlconfig: xml configuration string
        :type xmlconfig: :obj:`str`
        """
        self.tdwServer.XMLSettings = str(xmlconfig)
        self.tdwServer.OpenEntry()

    def record(self, jsondata):
        """ records one step

        :param jsondata: step JSON data
        :type jsondata: :obj:`str`
        """
        self.tdwServer.Record(jsondata)

    def closeEntry(self):
        """ closes the entry
        """
        self.tdwServer.CloseEntry()

    def closeFile(self):
        """ closes the file

        """
        self.tdwServer.CloseFile()


class OpenFile(Runner):

    """ OpenFile runner"""

    #: (:obj:`str`) command description
    description = "open a new H5 file"
    epilog = "" \
        + " examples:\n" \
        + "       nxsdata openfile /tmp/watertest.nxs \n" \
        + "       nxsdata openfile -s p02/tangodataserver/exp.01  " \
        + "/user/data/myfile.h5\n" \
        + "\n"

    def create(self):
        """ creates parser

        """
        parser = self._parser
        parser.add_argument(
            "-s", "--server", dest="server",
            help="writer server device name")

    def postauto(self):
        """ parser creator after autocomplete run """
        parser = self._parser
        parser.add_argument(
            'args', metavar='file_name', type=str, nargs='?',
            help='new newxus file name')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :returns: output information
        :rtype: :obj:`str`
        """
        tdwserver = NexusServer(options.server)
        if not options.args or len(options.args) < 1:
            self._parser.print_help()
            sys.exit(255)
        return tdwserver.openFile(options.args[0])


class SetData(Runner):

    """ SetData runner"""

    #: (:obj:`str`) command description
    description = "assign global JSON data"
    epilog = "" \
        + " examples:\n" \
        + "       nxsdata setdata ... \n" \
        + "\n"

    def create(self):
        """ creates parser

        """
        parser = self._parser
        parser.add_argument(
            "-s", "--server", dest="server",
            help="writer server device name"
        )
        parser.add_argument(
            'args', metavar='json_data_string', type=str, nargs='?',
            help='json data string')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :returns: output information
        :rtype: :obj:`str`
        """
        tdwserver = NexusServer(options.server)
        if not options.args or len(options.args) < 1:
            self._parser.print_help()
            sys.exit(255)
        return tdwserver.setData(options.args[0].strip())


class OpenEntry(Runner):

    """ OpenEntry runner"""

    #: (:obj:`str`) command description
    description = "create new entry"
    epilog = "" \
        + " examples:\n" \
        + "       nxsdata openentry ... \n" \
        + "\n"

    def create(self):
        """ creates parser

        """
        parser = self._parser
        parser.add_argument(
            "-s", "--server", dest="server",
            help="writer server device name")
        parser.add_argument(
            'args', metavar='xml_config', type=str, nargs='?',
            help='nexus writer configuration string')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :returns: output information
        :rtype: :obj:`str`
        """
        tdwserver = NexusServer(options.server)
        if not options.args or len(options.args) < 1:
            self._parser.print_help()
            sys.exit(255)
        return tdwserver.openEntry(options.args[0].strip())


class Record(Runner):

    """ Record runner"""

    #: (:obj:`str`) command description
    description = "record one step with step JSON data"
    epilog = "" \
        + " examples:\n" \
        + "       nxsdata record ... \n" \
        + "\n"

    def create(self):
        """ creates parser

        """
        parser = self._parser
        parser.add_argument(
            "-s", "--server", dest="server",
            help="writer server device name")
        parser.add_argument(
            'args', metavar='json_data_string', type=str, nargs='?',
            help='json data string')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :returns: output information
        :rtype: :obj:`str`
        """
        tdwserver = NexusServer(options.server)
        return tdwserver.record(
            options.args[0].strip() if options.args else '{}')


class CloseEntry(Runner):

    """ CloseEntry runner"""

    #: (:obj:`str`) command description
    description = "close the current entry"
    epilog = "" \
        + " examples:\n" \
        + "       nxsdata closeentry \n" \
        + "\n"

    def create(self):
        """ creates parser

        """
        parser = self._parser
        parser.add_argument(
            "-s", "--server", dest="server",
            help="writer server device name"
        )

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :returns: output information
        :rtype: :obj:`str`
        """
        tdwserver = NexusServer(options.server)
        return tdwserver.closeEntry()


class CloseFile(Runner):

    """ CloseFile runner"""

    #: (:obj:`str`) command description
    description = "close the current file"
    epilog = "" \
        + " examples:\n" \
        + "       nxsdata closefile \n" \
        + "\n"

    def create(self):
        """ creates parser

        """
        parser = self._parser
        parser.add_argument(
            "-s", "--server", dest="server",
            help="writer server device name"
        )

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :returns: output information
        :rtype: :obj:`str`
        """
        tdwserver = NexusServer(options.server)
        return tdwserver.closeFile()


class Servers(Runner):

    """ Servers runner"""

    #: (:obj:`str`) command description
    description = "get lists of tango data servers from " \
                  + "the current tango host"
    epilog = "" \
        + " examples:\n" \
        + "       nxsdata servers \n" \
        + "\n"

    def create(self):
        """ creates parser

        """
        parser = self._parser
        parser.add_argument(
            "-s", "--server", dest="server",
            help="tango host or writer server device name"
        )

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :returns: output information
        :rtype: :obj:`str`
        """
        return "\n".join(listServers(options.server, 'NXSConfigServer'))


def main():
    """ the main program function
    """

    #: pipe arguments
    pipe = ""
    if not sys.stdin.isatty():
        pp = sys.stdin.readlines()
        #: system pipe
        pipe = "".join(pp)

    description = "Command-line tool for writing NeXus files" \
                  + " with NXSDataWriter"

    epilog = 'For more help:\n  nxsdata <sub-command> -h'
    parser = NXSArgParser(
        description=description, epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.cmdrunners = [('openfile', OpenFile),
                         ('setdata', SetData),
                         ('openentry', OpenEntry),
                         ('record', Record),
                         ('closefile', CloseFile),
                         ('closeentry', CloseEntry)]
    runners = parser.createSubParsers()

    try:
        options = parser.parse_args()
    except ErrorException as e:
        sys.stderr.write("Error: %s\n" % str(e))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    if options.subparser is None:
        sys.stderr.write(
            "Error: %s\n" % str("too few arguments"))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    if options.subparser != 'servers':
        if not options.server:
            options.server = checkServer('NXSDataWriter')

        if not options.server:
            parser.subparsers[options.subparser].print_help()
            print("")
            sys.exit(255)

    #: command-line and pipe arguments
    parg = []
    if hasattr(options, "args"):
        parg = [options.args] if options.args else []
    if pipe:
        parg.append(pipe)
    options.args = parg

    result = runners[options.subparser].run(options)
    if result and str(result).strip():
        print(result)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" Command-line tool for ascess to the nexdatas configuration server """

import sys
import json
import xml.etree.ElementTree as et
from lxml.etree import XMLParser


if sys.version_info > (3,):
    unicode = str


def _parseString(text):
    if sys.version_info > (3,):
        return et.fromstring(
            bytes(text, "UTF-8"),
            parser=XMLParser(collect_ids=False))
    else:
        return et.fromstring(
            text,
            parser=XMLParser(collect_ids=False))


def _tostr(text):
    """ converts text  to str type

    :param text: text
    :type text: :obj:`bytes` or :obj:`unicode`
    :returns: text in str type
    :rtype: :obj:`str`
    """
    if isinstance(text, str):
        return text
    elif sys.version_info > (3,):
        return str(text, encoding="utf8")
    else:
        return str(text)


def _toxml(node):
    """ provides xml content of the whole node

    :param node: DOM node
    :type node: :class:`xml.dom.Node`
    :returns: xml content string
    :rtype: :obj:`str`
    """
    xml = _tostr(et.tostring(node, encoding='utf8', method='xml'))
    if xml.startswith("<?xml version='1.0' encoding='utf8'?>"):
        xml = str(xml[38:])
        return xml


class ParserTools(object):

    """ configuration server adapter
    """

    @classmethod
    def _getPureText(cls, node):
        """ collects text from text child nodes

        :param node: parent node
        :type node: :obj:`xml.etree.ElementTree.Element`
        """
        if node is not None:
            tnodes = ([node.text] if node.text else []) \
                     + [child.tail for child in node if child.tail]
            return unicode("".join(tnodes)).strip()
        return ""

    @classmethod
    def _getText(cls, node):
        """ provides  xml content of the node

        :param node: DOM node
        :type node: :class:`lxml.etree.Element`
        :returns: xml content string
        :rtype: :obj:`str`
        """
        if not node:
            return
        xmlc = _toxml(node)
        start = xmlc.find('>')
        end = xmlc.rfind('<')
        if start == -1 or end < start:
            return ""
        return xmlc[start + 1:end].replace("&lt;", "<").replace("&gt;", ">"). \
            replace("&quot;", "\"").replace("&amp;", "&")

    @classmethod
    def getRecord(cls, node):
        """ fetches record name or query from datasource node

        :param node: datasource node
        :type node: :class:`lxml.etree.Element`
        :returns: record name or query
        :rtype: :obj:`str`
        """
        withRec = ["CLIENT", "TANGO"]
        withQuery = ["DB"]
        if node.tag == 'datasource':
            dsource = node
        else:
            dsource = node.find(".//datasource")
        dstype = dsource.attrib["type"]
        res = ''
        if dstype and dstype in withRec:
            host = None
            port = None
            dname = None
            rname = None
            member = None
            device = dsource.findall("device")
            if device is not None and len(device) > 0:
                host = device[0].get("hostname")
                port = device[0].get("port")
                dname = device[0].get("name")
                member = device[0].get("member")
            surfix = ""
            prefix = ""
            if member or member != 'attribute':
                if member == 'property':
                    prefix = '@'
                elif member == 'command':
                    surfix = '()'
            record = dsource.findall("record")
            if record is not None and len(record) > 0:
                rname = record[0].get("name")
                if rname:
                    if dname:
                        if host:
                            if not port:
                                port = '10000'
                            res = '%s:%s/%s/%s%s%s' % (
                                host, port, dname, prefix, rname, surfix)
                        else:
                            res = '%s/%s%s%s' % (dname, prefix, rname, surfix)
                    else:
                        res = rname
                return res
        elif dstype and dstype in withQuery:
            query = dsource.find(".//query")
            if query and query.strip():
                return query.strip() or ""
        return res

    @classmethod
    def mergeDefinitions(cls, xmls):
        """ merges the xmls list of definitions xml strings
            to one output xml string

        :param xmls: a list of xml string with definitions
        :type xmls: :obj:`list` <:obj:`str`>
        :returns: one output xml string
        :rtype: :obj:`str`
        """
        rxml = ""
        if xmls:
            indom1 = _parseString(xmls[0])
            for xmlc in xmls[1:]:
                definition = _parseString(xmlc)
                if definition.tag == "definition":
                    definitions = [definition]
                else:
                    definitions = definition.findall("definition")
                for defin in definitions:
                    for tag in defin:
                        indom1.append(tag)
            rxml = _toxml(indom1)
        return rxml

    @classmethod
    def parseDataSources(cls, xmlc):
        """ provides datasources and its records from xml string

        :param xmlc: xml string
        :type xmlc: :obj:`str`
        :returns: list of datasource descriptions
        :rtype: :obj:`list` <:obj:`dict` <:obj:`str`, :obj:`str`>>
        """
        indom = _parseString(xmlc)
        return cls.__getDataSources(indom)

    @classmethod
    def __getDataSources(cls, node, direct=False):
        """ provides datasources and its records from xml string

        :param xmlc: xml string
        :type xmlc: :obj:`str`
        :returns: list of datasource descriptions
        :rtype: :obj:`list` <:obj:`dict` <:obj:`str`, :obj:`str`>>
        """
        dss = []
        if direct:
            dss.extend(node.findall("datasource"))
        else:
            dss.extend(node.findall(".//datasource"))
        dslist = []
        for ds in dss:
            if ds.tag == 'datasource':
                dstype = ds.attrib["type"]
                dsname = ds.attrib["name"]
                record = cls.getRecord(ds)
                dslist.append({
                    "source_type": dstype,
                    "source_name": dsname,
                    "source": record,
                })

        return dslist

    @classmethod
    def __getPath(cls, node):
        """ provides node path

        :param node: etree node
        :type node: :class:`lxml.etree.Element`
        :returns: node path
        :rtype: :obj:`str`
        """
        name = cls.__getAttr(node, "name")
        attr = False
        while node.getparent() is not None:
            onode = node
            node = node.getparent()
            if onode.tag == "attribute":
                attr = True
            else:
                attr = False
            if node.tag not in ["group", "field"]:
                return name
            else:
                gname = cls.__getAttr(node, "name")
                if not gname:
                    gname = cls.__getAttr(node, "type")
                    if len(gname) > 2:
                        gname = gname[2:]
                if attr:
                    name = gname + "@" + name
                else:
                    name = gname + "/" + name
                attr = False
        return name

    @classmethod
    def __getAttr(cls, node, name, tag=False):
        """ provides value of attirbute

        :param node: etree node
        :type node: :class:`lxml.etree.Element`
        :returns: attribute value
        :rtype: :obj:`str`
        """
        if name in node.attrib:
            return node.attrib[name]
        elif tag:
            if node.tag == "attribute":
                atnodes = [node]
            else:
                atnodes = []
            atnodes.extend(node.findall("attribute"))
            text = None
            for at in atnodes:
                if cls.__getAttr(at, "name") == name:
                    text = str(cls._getPureText(at)).strip()
                    if not text:
                        dss = cls.__getDataSources(at)
                        text = " ".join(["$datasources.%s" % ds for ds in dss])
            return text
        else:
            return None

    @classmethod
    def __getShape(cls, node):
        """ provides node shape

        :param node: etree node
        :type node: :class:`lxml.etree.Element`
        :returns: shape list
        :rtype: :obj:`list` <:obj:`int`>
        """
        rank = int(node.attrib["rank"])
        #        shape = ['*'] * rank
        shape = [None] * rank
        dims = node.findall("dim")
        for dim in dims:
            index = int(dim.attrib["index"])
            if "value" in dim.attrib:
                try:
                    value = int(dim.attrib["value"])
                except ValueError:
                    value = str(dim.attrib["value"])
                shape[index - 1] = value
            else:
                dss = dim.findall("datasource")
                if dss:
                    value = dss[0].get("name")
                    if not value:
                        value = '__unnamed__'
                    shape[index - 1] = "$datasources.%s" % value
                else:
                    tnodes = " ".join(
                        ([dim.text] if dim.text else []) +
                        [child.tail for child in dim if child.tail])
                    value = ("".join(tnodes)).strip()
                    try:
                        value = int(value)
                    except Exception:
                        value = value.strip()
                        if not value:
                            value = None
                    shape[index - 1] = value

        return shape

    @classmethod
    def __getChildrenByTagName(cls, parent, name):
        """ provides direct children by tag name

        :param parent: parent node
        :type parent: :class:`lxml.etree.Element`
        :param name: tag name
        :type name: :obj:`str`
        :returns: list of children
        :rtype: :obj:`list` <:class:`lxml.etree.Element`>
        """
        return [ch for ch in parent.findall(name) if ch.tag == name]

    @classmethod
    def parseFields(cls, xmlc):
        """ provides datasources and its records from xml string

        :param xmlc: xml string
        :type xmlc: :obj:`str`
        :returns: list of datasource descriptions
        :rtype: :obj:`list` < :obj:`dict` <:obj:`str`, `any`> >
        """
        tagname = "field"
        indom = _parseString(xmlc)
        nodes = []
        if indom.tag == tagname:
            nodes.append(indom)
        nodes.extend(indom.findall(".//%s" % tagname))
        taglist = []
        for nd in nodes:
            if nd.tag == tagname:
                nxtype = cls.__getAttr(nd, "type")
                units = cls.__getAttr(nd, "units")
                value = cls._getPureText(nd) or None
                trtype = cls.__getAttr(nd, "transformation_type", True)
                trvector = cls.__getAttr(nd, "vector", True)
                troffset = cls.__getAttr(nd, "offset", True)
                trdependson = cls.__getAttr(nd, "depends_on", True)
                nxpath = cls.__getPath(nd)
                dnodes = cls.__getChildrenByTagName(nd, "dimensions")
                shape = cls.__getShape(dnodes[0]) if dnodes else None
                stnodes = cls.__getChildrenByTagName(nd, "strategy")
                strategy = cls.__getAttr(stnodes[0], "mode") \
                    if stnodes else None

                sfdinfo = {
                    "strategy": strategy,
                    "nexus_path": nxpath,
                }
                fdinfo = {
                    "nexus_type": nxtype,
                    "units": units,
                    "shape": shape,
                    "trans_type": trtype,
                    "trans_vector": trvector,
                    "trans_offset": troffset,
                    "depends_on": trdependson,
                    "value": value
                }
                fdinfo.update(sfdinfo)
                dss = cls.__getDataSources(nd, direct=True)
                if dss:
                    for ds in dss:
                        ds.update(fdinfo)
                        taglist.append(ds)
                        nddss = cls.__getChildrenByTagName(nd, "datasource")
                        for ndds in nddss:
                            sdss = cls.__getDataSources(ndds, direct=True)
                            if sdss:
                                for sds in sdss:
                                    sds.update(sfdinfo)
                                    sds["source_name"] \
                                        = "\\" + sds["source_name"]
                                    taglist.append(sds)
                else:
                    taglist.append(fdinfo)

        return taglist

    @classmethod
    def parseAttributes(cls, xmlc):
        """ provides datasources and its records from xml string

        :param xmlc: xml string
        :type xmlc: :obj:`str`
        :returns: list of datasource descriptions
        :rtype: :obj:`list` < :obj:`dict` <:obj:`str`, `any`> >
        """
        tagname = "attribute"
        indom = _parseString(xmlc)
        nodes = []
        if indom.tag == tagname:
            nodes.append(indom)
        nodes.extend(indom.findall(".//%s" % tagname))
        taglist = []
        for nd in nodes:
            if nd.tag == tagname:

                nxtype = cls.__getAttr(nd, "type")
                units = cls.__getAttr(nd, "units")
                value = cls._getPureText(nd) or None
                trtype = cls.__getAttr(nd, "transformation_type", True)
                trvector = cls.__getAttr(nd, "vector", True)
                troffset = cls.__getAttr(nd, "offset", True)
                trdependson = cls.__getAttr(nd, "depends_on", True)
                nxpath = cls.__getPath(nd)
                dnodes = cls.__getChildrenByTagName(nd, "dimensions")
                shape = cls.__getShape(dnodes[0]) if dnodes else None
                stnodes = cls.__getChildrenByTagName(nd, "strategy")
                strategy = cls.__getAttr(stnodes[0], "mode") \
                    if stnodes else None
                sfdinfo = {
                    "strategy": strategy,
                    "nexus_path": nxpath,
                }
                fdinfo = {
                    "nexus_type": nxtype,
                    "units": units,
                    "shape": shape,
                    "trans_type": trtype,
                    "trans_vector": trvector,
                    "trans_offset": troffset,
                    "depends_on": trdependson,
                    "value": value
                }
                fdinfo.update(sfdinfo)
                dss = cls.__getDataSources(nd, direct=True)
                if dss:
                    for ds in dss:
                        ds.update(fdinfo)
                        taglist.append(ds)
                        nddss = cls.__getChildrenByTagName(nd, "datasource")
                        for ndds in nddss:
                            sdss = cls.__getDataSources(ndds, direct=True)
                            if sdss:
                                for sds in sdss:
                                    sds.update(sfdinfo)
                                    sds["source_name"] \
                                        = "\\" + sds["source_name"]
                                    taglist.append(sds)
                else:
                    taglist.append(fdinfo)

        return taglist

    @classmethod
    def parseLinks(cls, xmlc):
        """ provides datasources and its records from xml string

        :param xmlc: xml string
        :type xmlc: :obj:`str`
        :returns: list of datasource descriptions
        :rtype: :obj:`list` < :obj:`dict` <:obj:`str`, `any`> >
        """
        tagname = "link"
        indom = _parseString(xmlc)
        nodes = []
        if indom.tag == tagname:
            nodes.append(indom)
        nodes.extend(indom.findall(".//%s" % tagname))
        taglist = []
        for nd in nodes:
            if nd.tag == tagname:

                target = cls.__getAttr(nd, "target")
                value = cls._getPureText(nd) or None
                nxpath = cls.__getPath(nd)
                stnodes = cls.__getChildrenByTagName(nd, "strategy")
                strategy = cls.__getAttr(stnodes[0], "mode") \
                    if stnodes else None

                sfdinfo = {
                    "strategy": strategy,
                    "nexus_path": "[%s]" % nxpath,
                }
                fdinfo = {
                    "value": value
                }
                fdinfo.update(sfdinfo)
                dss = cls.__getDataSources(nd, direct=True)
                if dss:
                    for ds in dss:
                        ds.update(fdinfo)
                        taglist.append(ds)
                        nddss = cls.__getChildrenByTagName(nd, "datasource")
                        for ndds in nddss:
                            sdss = cls.__getDataSources(ndds, direct=True)
                            if sdss:
                                for sds in sdss:
                                    sds.update(sfdinfo)
                                    sds["source_name"] \
                                        = "\\" + sds["source_name"]
                                    taglist.append(sds)
                else:
                    taglist.append(fdinfo)
                    if target.strip():
                        fdinfo2 = dict(fdinfo)
                        fdinfo2["nexus_path"] = "\\-> %s" % target
                        taglist.append(fdinfo2)

        return taglist

    @classmethod
    def parseRecord(cls, xmlc):
        """ provides source record from xml string

        :param xmlc: xml string
        :type xmlc: :obj:`str`
        :returns: source record
        :rtype: :obj:`str`
        """
        indom = _parseString(xmlc)
        return cls.getRecord(indom)


class TableTools(object):

    """ configuration server adapter
    """

    def __init__(self, description, nonone=None):
        """ constructor

        :param description: description list
        :type description:  :obj:`list` <:obj:`str`>
        :param nonone: list of parameters which have to exist to be shown
        :type nonone:  :obj:`list` <:obj:`str`>
        """
        #: (:obj:`list` <:obj:`str`>)
        #:    list of parameters which have to exist to be shown
        self.__nonone = nonone or []
        #: (:obj:`list` <:obj:`str`>)
        #:    description list
        self.__description = []
        #: (:obj:`dict` <:obj:`str` , :obj:`int`>) header sizes
        self.__hdsizes = {}
        #: (:obj:`list` <:obj:`str`>) table headers
        self.headers = [
            'nexus_path',
            'nexus_type',
            'strategy',
            'shape',
            'units',
            'depends_on',
            'trans_type',
            'trans_vector',
            'trans_offset',
            'source_name',
            'source_type',
            'source',
            'value',
        ]
        #: (:obj:`str`) table title
        self.title = None
        self.__loadDescription(description)

    def __loadDescription(self, description):
        """ loads description

        :param description:  description list
        :type description:  :obj:`list` <:obj:`str`>
        """
        for desc in description:
            if desc is None:
                self.__description.append(desc)
                continue
            skip = False
            field = desc.get("nexus_path", "").split('/')[-1]
            value = desc.get("value", "")
            if field == 'depends_on' and value:
                desc["depends_on"] = "[%s]" % value
            for hd in self.__nonone:
                vl = desc.get(hd, "")
                if isinstance(vl, (list, tuple)):
                    vl = self.__toString(vl)
                if not vl:
                    skip = True
                    break
            if not skip:
                self.__description.append(desc)

        for desc in self.__description:
            if desc is None:
                continue
            for hd, vl in desc.items():
                if hd not in self.__nonone or vl:
                    if hd not in self.__hdsizes.keys():
                        self.__hdsizes[hd] = max(len(hd) + 1, 5)
                    if isinstance(vl, (list, tuple)):
                        vl = self.__toString(vl)
                    if not isinstance(vl, str):
                        vl = str(vl)
                    if self.__hdsizes[hd] <= len(vl):
                        self.__hdsizes[hd] = len(vl) + 1

    @classmethod
    def __toString(cls, lst):
        """ converts list to string

        :param lst: given list
        :type lst: :obj:`list` <:obj:`str`>
        :returns: list in string representation
        :rtype: :obj:`str`
        """
        res = []
        for it in lst:
            res.append(it or "*")
        return str(res)

    def generateList(self):
        """ generate row lists of table

        :returns:  table rows
        :rtype: :obj:`list` <:obj:`str`>
        """
        lst = [""]
        if self.title is not None:
            lst.append(self.title)
            lst.append("-" * len(self.title))
            lst.append("")

        headers = [hd for hd in self.headers if hd in self.__hdsizes.keys()]
        line = ""
        for hd in headers:
            line += "=" * (self.__hdsizes[hd] - 1) + " "
        lst.append(line)

        line = ""
        for hd in headers:
            line += hd + " " * (self.__hdsizes[hd] - len(hd))
        lst.append(line)

        line = ""
        for hd in headers:
            line += "=" * (self.__hdsizes[hd] - 1) + " "
        lst.append(line)

        for desc in self.__description:
            line = ""
            if desc is None:
                for hd in headers:
                    line += "=" * (self.__hdsizes[hd] - 1) + " "
                lst.append(line.rstrip())
                continue
            line = ""
            for hd in headers:
                vl = desc[hd] if hd in desc else None
                if isinstance(vl, (list, tuple)):
                    vl = self.__toString(vl)
                elif vl is None:
                    vl = ""
                elif not isinstance(vl, str):
                    vl = str(vl)
                line += vl + " " * (self.__hdsizes[hd] - len(vl))

            lst.append(line.rstrip())

        line = ""
        for hd in headers:
            line += "=" * (self.__hdsizes[hd] - 1) + " "
        lst.append(line)

        lst.append("")

        return lst


class TableDictTools(object):

    """ configuration server adapter
    """

    def __init__(self, description, nonone=None):
        """ constructor

        :param description: description list
        :type description:  :obj:`list` <:obj:`str`>
        :param nonone: list of parameters which have to exist to be shown
        :type nonone:  :obj:`list` <:obj:`str`>
        """
        #: (:obj:`list` <:obj:`str`>)
        #:    description list
        self.__description = description

        #: (:obj:`list` <:obj:`str`>)
        #:    headers
        self.headers = [
            'Timer',
            'DataSourceSelection',
            'ComponentSelection',
            'ComponentPreselection',
            'DataSourcePreselection',

            'UserData',
            'AppendEntry',
            'ConfigDevice',
            'WriterDevice',
            'Door',

            'DynamicComponents',
            'ComponentsFromMntGrp',
            'DefaultDynamicLinks',
            'DefaultDynamicPath',
            'UnplottedComponents',
            'OptionalComponents',
            'ConfigVariables',

            # 'TimeZone',
            # 'Version',
            # 'OrderedChannels',
            # 'PreselectingDataSources',
            # 'MntGrpConfiguration',
            # 'ChannelProperties',
            # 'MntGrp',
        ]

        self.headertypenames = {
            'MntGrp': ('str', None),
            'Timer': ('list', 'Timer(s)'),
            'ComponentSelection':
            ('tdict', 'Detector Components'),
            'DataSourceSelection': (
                'tdict', 'Pool/Dynamic Detector Components'),
            'ComponentPreselection': (
                'tdict', 'Descriptive Components'),
            'DataSourcePreselection': (
                'tdict', 'Descriptive Dynamic Components'),
            'UserData': ('dict', 'User Data'),
            'AppendEntry': ('str', None),
            'ConfigDevice': ('str', None),
            'WriterDevice': ('str', None),
            'UnplottedComponents': ('list', 'Unplotted Components'),
            'MntGrpConfiguration': ('str', None),
            'Version': ('str', None),
            'TimeZone': ('str', None),
            'Door': ('str', None),
            'DynamicComponents': ('str', None),
            'PreselectingDataSources': ('list', None),
            'OrderedChannels': ('list', None),
            'DefaultDynamicLinks': ('str', None),
            'ConfigVariables': ('str', None),
            'ComponentsFromMntGrp': ('str', None),
            'ChannelProperties': ('str', None),
            'OptionalComponents': ('list', None),
            'DefaultDynamicPath': ('str', None),
        }

        self.typemethods = {
            'str': self._getstr,
            'list': self._getlist,
            'tdict': self._gettdict,
            'dict': self._getdict,
        }

        #: (:obj:`str`) table title
        self.title = None
        self.maxnamesize = max(
            [len(hd) for hd in self.headers] +
            [len(self.headertypenames[hd][1])
             for hd in self.headers
             if self.headertypenames[hd][1]]
        )

    def _getstr(self, name, value):
        space = " " * (self.maxnamesize - len(name))
        sep = ":" if name != " " else name
        return ["%s%s %s%s" % (name, sep, space, value)]

    def _getlist(self, name, value):
        space = " " * (self.maxnamesize - len(name))
        svalue = ", ".join(json.loads(value)) if value else ""
        sep = ":" if name != " " else name
        return ["%s%s %s%s" % (name, sep, space, svalue)]

    def _gettdict(self, name, value):
        space = " " * (self.maxnamesize - len(name))
        dvl = json.loads(value)
        svalue = ""
        if dvl:
            svalue = ", ".join([key for key in dvl.keys() if dvl[key]])
        sep = ":" if name != " " else name
        return ["%s%s %s%s" % (name, sep, space, svalue)]

    def _getdict(self, name, value):
        space = " " * (self.maxnamesize - len(name))
        sep = ":" if name != " " else name
        return ["%s%s %s%s" % (name, sep, space, value)]

    @classmethod
    def __toString(cls, lst):
        """ converts list to string

        :param lst: given list
        :type lst: :obj:`list` <:obj:`str`>
        :returns: list in string representation
        :rtype: :obj:`str`
        """
        res = []
        for it in lst:
            res.append(it or "*")
        return str(res)

    def generateList(self):
        """ generate row lists of table

        :returns:  table rows
        :rtype: :obj:`list` <:obj:`str`>
        """
        self.maxnamesize = max(
            [len(hd) for hd in self.headers] +
            [len(self.headertypenames[hd][1])
             for hd in self.headers
             if self.headertypenames[hd][1]]
        )
        lst = [""]

        if self.title is not None:
            lst.append(self.title)
            lst.append("-" * len(self.title))
            lst.append("")

        for desc in self.__description:
            for hd in self.headers:
                if hd in desc.keys():
                    htp, name = self.headertypenames[hd]
                    method = self.typemethods[htp]
                    lst.extend(method(name or hd, desc[hd]))
        return lst



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" Command-line tool for showing meta data from Nexus Files"""

import sys
import argparse

from .nxsparser import TableTools
from .nxsfileparser import NXSFileParser
from .nxsargparser import (Runner, NXSArgParser, ErrorException)
from . import filewriter


WRITERS = {}
try:
    from . import pniwriter
    WRITERS["pni"] = pniwriter
except Exception:
    pass
try:
    from . import h5pywriter
    WRITERS["h5py"] = h5pywriter
except Exception:
    pass

try:
    from . import h5cppwriter
    WRITERS["h5cpp"] = h5cppwriter
except Exception:
    pass


class General(Runner):

    """ General runner"""

    #: (:obj:`str`) command description
    description = "show general information for the nexus file"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxsfileinfo general /user/data/myfile.nxs\n" \
        + "\n"

    def create(self):
        """ creates parser

        """
        self._parser.add_argument(
            "--pni", action="store_true",
            default=False, dest="pni",
            help="use pni module as a nexus reader")
        self._parser.add_argument(
            "--h5py", action="store_true",
            default=False, dest="h5py",
            help="use h5py module as a nexus reader")
        self._parser.add_argument(
            "--h5cpp", action="store_true",
            default=False, dest="h5cpp",
            help="use h5cpp module as a nexus reader")

    def postauto(self):
        """ parser creator after autocomplete run """
        self._parser.add_argument(
            'args', metavar='nexus_file', type=str, nargs=1,
            help='new nexus file name')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :returns: output information
        :rtype: :obj:`str`
        """
        if options.h5cpp:
            writer = "h5cpp"
        elif options.h5py:
            writer = "h5py"
        elif options.pni:
            writer = "pni"
        elif "h5cpp" in WRITERS.keys():
            writer = "h5cpp"
        elif "h5py" in WRITERS.keys():
            writer = "h5py"
        else:
            writer = "pni"
        if (options.pni and options.h5py and options.h5cpp) or \
           writer not in WRITERS.keys():
            sys.stderr.write("nxsfileinfo: Writer '%s' cannot be opened\n"
                             % writer)
            sys.stderr.flush()
            self._parser.print_help()
            sys.exit(255)
        wrmodule = WRITERS[writer.lower()]
        try:
            fl = filewriter.open_file(
                options.args[0], readonly=True,
                writer=wrmodule)
        except Exception:
            sys.stderr.write("nxsfileinfo: File '%s' cannot be opened\n"
                             % options.args[0])
            sys.stderr.flush()
            self._parser.print_help()
            sys.exit(255)

        root = fl.root()
        self.show(root)
        fl.close()

    @classmethod
    def parseentry(cls, entry, description):
        """ parse entry of nexus file

        :param entry: nexus entry node
        :type entry: :class:`pni.io.nx.h5.nxgroup`
        :param description: dict description list
        :type description: :obj:`list` <:obj:`dict` <:obj:`str`, `any` > >
        :return: (key, value) name pair of table headers
        :rtype: [:obj:`str`, :obj:`str`]

        """
        key = "A"
        value = "B"
        at = None
        try:
            at = entry.attributes["NX_class"]
        except Exception:
            pass
        if at and at[...] == 'NXentry':
            # description.append(None)
            # value = filewriter.first(value)
            key = "Scan entry:"
            value = entry.name
            # description.append({key: "Scan entry:", value: entry.name})
            # description.append(None)
            try:
                vl = filewriter.first(entry.open("title")[...])
                description.append(
                    {key: "Title:", value: vl})
            except Exception:
                sys.stderr.write("nxsfileinfo: title cannot be found\n")
                sys.stderr.flush()
            try:
                vl = filewriter.first(entry.open("experiment_identifier")[...])
                description.append(
                    {key: "Experiment identifier:",
                     value: vl})
            except Exception:
                sys.stderr.write(
                    "nxsfileinfo: experiment identifier cannot be found\n")
                sys.stderr.flush()
            for ins in entry:
                if isinstance(ins, filewriter.FTGroup):
                    iat = ins.attributes["NX_class"]
                    if iat and iat[...] == 'NXinstrument':
                        try:
                            vl = filewriter.first(ins.open("name")[...])
                            description.append({
                                key: "Instrument name:",
                                value: vl})
                        except Exception:
                            sys.stderr.write(
                                "nxsfileinfo: instrument name cannot "
                                "be found\n")
                            sys.stderr.flush()
                        try:
                            vl = filewriter.first(
                                ins.open("name").attributes["short_name"][...])
                            description.append({
                                key: "Instrument short name:",
                                value: vl
                            })
                        except Exception:
                            sys.stderr.write(
                                "nxsfileinfo: instrument short name cannot"
                                " be found\n")
                            sys.stderr.flush()

                        for sr in ins:
                            if isinstance(sr, filewriter.FTGroup):
                                sat = sr.attributes["NX_class"]
                                if sat and sat[...] == 'NXsource':
                                    try:
                                        vl = filewriter.first(
                                            sr.open("name")[...])
                                        description.append({
                                            key: "Source name:",
                                            value: vl})
                                    except Exception:
                                        sys.stderr.write(
                                            "nxsfileinfo: source name"
                                            " cannot be found\n")
                                        sys.stderr.flush()
                                    try:
                                        vl = filewriter.first(
                                            sr.open("name").attributes[
                                                "short_name"][...])
                                        description.append({
                                            key: "Source short name:",
                                            value: vl})
                                    except Exception:
                                        sys.stderr.write(
                                            "nxsfileinfo: source short name"
                                            " cannot be found\n")
                                        sys.stderr.flush()
                    elif iat and iat[...] == 'NXsample':
                        try:
                            vl = filewriter.first(ins.open("name")[...])
                            description.append({
                                key: "Sample name:",
                                value: vl})
                        except Exception:
                            sys.stderr.write(
                                "nxsfileinfo: sample name cannot be found\n")
                            sys.stderr.flush()
                        try:
                            vl = filewriter.first(
                                ins.open("chemical_formula")[...])
                            description.append({
                                key: "Sample formula:",
                                value: vl})
                        except Exception:
                            sys.stderr.write(
                                "nxsfileinfo: sample formula cannot"
                                " be found\n")
                            sys.stderr.flush()
            try:
                vl = filewriter.first(entry.open("start_time")[...])
                description.append({key: "Start time:", value: vl})
            except Exception:
                sys.stderr.write("nxsfileinfo: start time cannot be found\n")
                sys.stderr.flush()
            try:
                vl = filewriter.first(entry.open("end_time")[...])
                description.append({key: "End time:",
                                    value: vl})
            except Exception:
                sys.stderr.write("nxsfileinfo: end time cannot be found\n")
                sys.stderr.flush()
            if "program_name" in entry.names():
                pn = entry.open("program_name")
                pname = filewriter.first(pn.read())
                attr = pn.attributes
                names = [att.name for att in attr]
                if "scan_command" in names:
                    scommand = filewriter.first(attr["scan_command"][...])
                    pname = "%s (%s)" % (pname, scommand)
                description.append({key: "Program:", value: pname})
        return [key, value]

    def show(self, root):
        """ show general informations

        :param root: nexus file root
        :type root: class:`pni.io.nx.h5.nxroot`
        """

        description = []

        attr = root.attributes

        names = [at.name for at in attr]
        fname = filewriter.first(
            (attr["file_name"][...]
             if "file_name" in names else " ") or " ")
        title = "File name: '%s'" % fname

        print("")
        for en in root:
            description = []
            headers = self.parseentry(en, description)
            ttools = TableTools(description)
            ttools.title = title
            ttools.headers = headers
            rstdescription = ttools.generateList()
            title = ""
            print("\n".join(rstdescription).strip())
            print("")


class Field(Runner):

    """ Field runner"""

    #: (:obj:`str`) command description
    description = "show field information for the nexus file"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxsfileinfo field /user/data/myfile.nxs\n" \
        + "       nxsfileinfo field /user/data/myfile.nxs -g\n" \
        + "       nxsfileinfo field /user/data/myfile.nxs -s\n" \
        + "\n"

    def create(self):
        """ creates parser

        """
        self._parser.add_argument(
            "-c", "--columns",
            help="names of column to be shown (separated by commas "
            "without spaces). The possible names are: "
            "depends_on, dtype, full_path, nexus_path, nexus_type, shape,"
            " source, source_name, source_type, strategy, trans_type, "
            "trans_offset, trans_vector, units, value",
            dest="headers", default="")
        self._parser.add_argument(
            "-f", "--filters",
            help="full_path filters (separated by commas "
            "without spaces). Default: '*'. E.g. '*:NXsample/*'",
            dest="filters", default="")
        self._parser.add_argument(
            "-v", "--values",
            help="field names which value should be stored"
            " (separated by commas "
            "without spaces). Default: depends_on",
            dest="values", default="")
        self._parser.add_argument(
            "-g", "--geometry", action="store_true",
            default=False, dest="geometry",
            help="perform geometry full_path filters, i.e."
            "*:NXtransformations/*,*/depends_on. "
            "It works only when  -f is not defined")
        self._parser.add_argument(
            "-s", "--source", action="store_true",
            default=False, dest="source",
            help="show datasource parameters")
        self._parser.add_argument(
            "--pni", action="store_true",
            default=False, dest="pni",
            help="use pni module as a nexus reader")
        self._parser.add_argument(
            "--h5py", action="store_true",
            default=False, dest="h5py",
            help="use h5py module as a nexus reader")
        self._parser.add_argument(
            "--h5cpp", action="store_true",
            default=False, dest="h5cpp",
            help="use h5cpp module as a nexus reader")

    def postauto(self):
        """ parser creator after autocomplete run """
        self._parser.add_argument(
            'args', metavar='nexus_file', type=str, nargs=1,
            help='new nexus file name')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :returns: output information
        :rtype: :obj:`str`
        """
        if options.h5cpp:
            writer = "h5cpp"
        elif options.h5py:
            writer = "h5py"
        elif options.pni:
            writer = "pni"
        elif "h5cpp" in WRITERS.keys():
            writer = "h5cpp"
        elif "h5py" in WRITERS.keys():
            writer = "h5py"
        else:
            writer = "pni"
        if (options.pni and options.h5py) or writer not in WRITERS.keys():
            sys.stderr.write("nxsfileinfo: Writer '%s' cannot be opened\n"
                             % writer)
            sys.stderr.flush()
            self._parser.print_help()
            sys.exit(255)
        wrmodule = WRITERS[writer.lower()]
        try:
            fl = filewriter.open_file(
                options.args[0], readonly=True,
                writer=wrmodule)
        except Exception:
            sys.stderr.write("nxsfileinfo: File '%s' cannot be opened\n"
                             % options.args[0])
            sys.stderr.flush()
            self._parser.print_help()
            sys.exit(255)

        root = fl.root()
        self.show(root, options)
        fl.close()

    def show(self, root, options):
        """ the main function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :param root: nexus file root
        :type root: class:`pni.io.nx.h5.nxroot`
        """
        #: (:obj:`list`< :obj:`str`>)   \
        #     parameters which have to exists to be shown
        toshow = None

        #: (:obj:`list`< :obj:`str`>)  full_path filters
        filters = []

        #: (:obj:`list`< :obj:`str`>)  column headers
        headers = ["nexus_path", "source_name", "units",
                   "dtype", "shape", "value"]
        if options.geometry:
            filters = ["*:NXtransformations/*", "*/depends_on"]
            headers = ["nexus_path", "source_name", "units",
                       "trans_type", "trans_vector", "trans_offset",
                       "depends_on"]
        if options.source:
            headers = ["source_name", "nexus_type", "shape", "strategy",
                       "source"]
            toshow = ["source_name"]
        #: (:obj:`list`< :obj:`str`>)  field names which value should be stored
        values = ["depends_on"]

        if options.headers:
            headers = options.headers.split(',')
        if options.filters:
            filters = options.filters.split(',')
        if options.values:
            values = options.values.split(',')

        nxsparser = NXSFileParser(root)
        nxsparser.filters = filters
        nxsparser.valuestostore = values
        nxsparser.parse()

        description = []
        ttools = TableTools(nxsparser.description, toshow)
        ttools.title = "File name: '%s'" % options.args[0]
        ttools.headers = headers
        description.extend(ttools.generateList())
        print("\n".join(description))


def main():
    """ the main program function
    """

    description = "Command-line tool for showing meta data" \
                  + " from Nexus Files"

    epilog = 'For more help:\n  nxsfileinfo <sub-command> -h'
    parser = NXSArgParser(
        description=description, epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.cmdrunners = [('field', Field),
                         ('general', General)]
    runners = parser.createSubParsers()

    try:
        options = parser.parse_args()
    except ErrorException as e:
        sys.stderr.write("Error: %s\n" % str(e))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    if options.subparser is None:
        sys.stderr.write(
            "Error: %s\n" % str("too few arguments"))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    result = runners[options.subparser].run(options)
    if result and str(result).strip():
        print(result)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" NeXus tool argumen parser
"""

import argparse
import argcomplete


class Runner(object):

    """ abstract runner"""

    #: (:obj:`str`) command description
    description = "abstract runner"

    #: (:obj:`str`) command epilog
    epilog = None

    def __init__(self, parser):
        """ parser creator

        :param parser: option parser
        :type parser: :class:`NXSFileInfoArgParser`
        """
        #: (:class:`NXSFileInfoArgParser`) option parser
        self._parser = parser

    def create(self):
        """ parser creator """

    def postauto(self):
        """ parser creator after autocomplete run """

    def run(self, options):
        """ run commandthe main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """


class ErrorException(Exception):

    """ error parser exception """
    pass


class NXSArgParser(argparse.ArgumentParser):

    """ Argument parser with error exception
    """

    def __init__(self, **kwargs):
        """ constructor

        :param kwargs: :class:`argparse.ArgumentParser`
                       parameter dictionary
        :type kwargs: :obj: `dict` <:obj:`str`, `any`>
        """
        argparse.ArgumentParser.__init__(self, **kwargs)
        self.subparsers = {}
        #: (:obj:`dict` <:obj:`str`, :class:`Runner`>) \
        #    nxsfileinfo sub-command classes
        self.cmdrunners = []

    def error(self, message):
        """ error handler

        :param message: error message
        :type message: :obj:`str`
        """
        raise ErrorException(message)

    def createSubParsers(self):
        """ creates command-line parameters parser

        :returns: command runner
        :rtype: :class:`Runner`
        """

        pars = {}
        subparsers = self.add_subparsers(
            help='sub-command help', dest="subparser")

        for cmd, klass in self.cmdrunners:
            self.subparsers[cmd] = subparsers.add_parser(
                cmd, help='%s' % klass.description,
                description=klass.description,
                epilog=klass.epilog,
                formatter_class=argparse.RawDescriptionHelpFormatter
            )
            pars[cmd] = klass(self.subparsers[cmd])
            pars[cmd].create()

        argcomplete.autocomplete(self)

        for cmd, klass in self.cmdrunners:
            pars[cmd].postauto()

        return pars



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" Command-line tool to merge images of external file-formats
into the master NeXus file
"""

import sys
import os
import re
import shutil
import fabio
import signal
import argparse
import numpy
import json

from .filenamegenerator import FilenameGenerator
from .nxsargparser import (Runner, NXSArgParser, ErrorException)
from . import filewriter


if sys.version_info > (3,):
    unicode = str
    long = int
else:
    bytes = str


WRITERS = {}
try:
    from . import pniwriter
    WRITERS["pni"] = pniwriter
except Exception:
    pass

try:
    from . import h5pywriter
    WRITERS["h5py"] = h5pywriter
except Exception:
    pass

try:
    from . import h5cppwriter
    WRITERS["h5cpp"] = h5cppwriter
except Exception:
    pass


def getcompression(compression):
    """ converts compression string to a deflate level parameter
        or list with [filterid, opt1, opt2, ...]

    :param compression: compression string
    :type compression: :obj:`str`
    :returns: deflate level parameter
              or list with [filterid, opt1, opt2, ...]
    :rtype: :obj:`int` or :obj:`list` < :obj:`int` > or `None`

    """
    if compression:
        if isinstance(compression, int) or ":" not in compression:
            level = None
            try:
                level = int(compression)
            except Exception:
                raise Exception(
                    "Error: argument -c/--compression: "
                    "invalid int value: '%s'\n" % compression)
            return level
        else:
            opts = None
            try:
                sfid, sopts = compression.split(":")
                opts = [int(sfid)]
                opts.extend([int(opt) for opt in sopts.split(",")])
            except Exception:
                raise Exception(
                    "Error: argument -c/--compression: "
                    "invalid format: '%s'\n" % compression)
            return opts
        return


class Collector(object):

    """ Collector merge images of external file-formats
    into the master NeXus file
    """

    def __init__(self, nexusfilename, compression=2,
                 skipmissing=False, storeold=False, testmode=False,
                 writer=None):
        """ The constructor creates the collector object

        :param nexusfilename: the nexus file name
        :type nexusfilename: :obj:`str`
        :param compression: compression rate
        :type compression: :obj:`int`
        :param skipmissing: if skip missing images
        :type skipmissing: :obj:`bool`
        :param storeold: if backup the input file
        :type storeold: :obj:`bool`
        :param testmode: if run in a test mode
        :type testmode: :obj:`bool`
        """
        self.__nexusfilename = nexusfilename
        self.__compression = compression
        self.__skipmissing = skipmissing
        self.__testmode = testmode
        self.__storeold = storeold
        self.__tempfilename = None
        self.__filepattern = re.compile("[^:]+:\\d+:\\d+")
        self.__nxsfile = None
        self.__break = False
        self.__fullfilename = None
        self.__wrmodule = None
        if writer and writer.lower() in WRITERS.keys():
            self.__wrmodule = WRITERS[writer.lower()]
        self.__siginfo = dict(
            (signal.__dict__[sname], sname)
            for sname in ('SIGINT', 'SIGHUP', 'SIGALRM', 'SIGTERM'))

        for sig in self.__siginfo.keys():
            signal.signal(sig, self._signalhandler)

    def _signalhandler(self, sig, _):
        """ signal handler

        :param sig: signal name, i.e. 'SIGINT', 'SIGHUP', 'SIGALRM', 'SIGTERM'
        :type sig: :obj:`str`
        """
        if sig in self.__siginfo.keys():
            self.__break = True
            print("terminated by %s" % self.__siginfo[sig])

    def _createtmpfile(self):
        """ creates temporary file
        """
        self.__tempfilename = self.__nexusfilename + ".__nxscollect_temp__"
        while os.path.exists(self.__tempfilename):
            self.__tempfilename += "_"
        shutil.copy2(self.__nexusfilename, self.__tempfilename)

    def _storeoldfile(self):
        """ makes back up of the input file
        """
        temp = self.__nexusfilename + ".__nxscollect_old__"
        while os.path.exists(temp):
            temp += "_"
        shutil.move(self.__nexusfilename, temp)

    def _filegenerator(self, filestr):
        """ provides file name generator from file string

        :param filestr: file string
        :type: filestr: :obj:`str`
        :returns: file name generator or a list of file names
        :rtype: :class:`methodinstance`
        """
        if self.__filepattern.match(filestr):
            return FilenameGenerator.from_slice(filestr)
        else:
            def _files():
                return [filestr]
            return _files

    @classmethod
    def _absolutefilename(cls, filename, masterfile):
        """ provides absolute image file name

        :param filename: image file name
        :type: filename: :obj:`str`
        :param masterfile: nexus file name
        :type: masterfile: :obj:`str`
        :returns: absolute image file name
        :rtype: :obj:`str`
        """
        if not os.path.isabs(filename):
            nexusfilepath = os.path.join('/', *os.path.abspath(
                masterfile).split('/')[:-1])
            filename = os.path.abspath(os.path.join(nexusfilepath, filename))
        return filename

    def _findfile(self, filename, nname=None):
        """ searches for absolute image file name

        :param filename: image file name
        :type: filename: :obj:`str`
        :param nname: hdf5 node name
        :typ nname: :obj:`str`

        :returns: absolute image file name
        :rtype: :obj:`str`
        """
        filelist = []

        if nname is not None:
            tmpfname = '%s/%s/%s' % (
                os.path.splitext(self.__nexusfilename)[0],
                nname,
                filename.split("/")[-1])
            if os.path.exists(tmpfname):
                return tmpfname
            else:
                filelist.append(tmpfname)
            tmpfname = '%s/%s/%s' % (
                os.path.splitext(self.__fullfilename)[0],
                nname,
                filename.split("/")[-1])
            if os.path.exists(tmpfname):
                return tmpfname
            else:
                filelist.append(tmpfname)
        tmpfname = self._absolutefilename(filename, self.__nexusfilename)
        if os.path.exists(tmpfname):
            return tmpfname
        else:
            filelist.append(tmpfname)
        tmpfname = self._absolutefilename(filename, self.__fullfilename)
        if os.path.exists(tmpfname):
            return tmpfname
        else:
            filelist.append(tmpfname)
        if os.path.exists(filename):
            return filename
        else:
            filelist.append(filename)
        if not self.__skipmissing:
            raise Exception(
                "Cannot open any of %s files" % sorted(set(filelist)))
        else:
            print("Cannot open any of %s files" % sorted(set(filelist)))
        return None

    def _loadrawimage(self, filename, dtype, shape=None):
        """ loads image from file

        :param filename: image file name
        :type filename: :obj:`str`
        :param dtype: field data type
        :type dtype: :obj:`str`
        :param shape: field shape
        :type shape: :obj:`list` <:obj:`int` >
        :returns: (image data, image data type, image shape)
        :rtype: (:class:`numpy.ndarray`, :obj:`str`, :obj:`list` <:obj:`int`>)
        """
        try:
            idata = None

            fl = open(filename, "rb")
            idata = numpy.fromfile(fl, dtype=dtype)
            if shape:
                idata = idata.reshape(shape)
            dtype = idata.dtype.__str__()
            shape = idata.shape
            if idata is not None:
                return idata, dtype, shape
            else:
                raise Exception("Cannot open a file %s" % filename)
        except Exception as e:
            print(str(e))
            if not self.__skipmissing:
                raise Exception("Cannot open a file %s" % filename)
            else:
                print("Cannot open a file %s" % filename)

            return None, None, None

    def _loadimage(self, filename):
        """ loads image from file

        :param filename: image file name
        :type filename: :obj:`str`
        :returns: (image data, image data type, image shape)
        :rtype: (:class:`numpy.ndarray`, :obj:`str`, :obj:`list` <:obj:`int`>)
        """
        try:
            dtype = None
            shape = None
            idata = None
            image = fabio.open(filename)
            if image:
                idata = image.data[...]
                dtype = image.data.dtype.__str__()
                shape = image.data.shape
                return idata, dtype, shape
            else:
                raise Exception("Cannot open a file %s" % filename)
        except Exception:
            if not self.__skipmissing:
                raise Exception("Cannot open a file %s" % filename)
            else:
                print("Cannot open a file %s" % filename)

            return None, None, None

    def _loadh5data(self, filename):
        """ loads image from hdf5 file

        :param filename: hdf5 image file name
        :type filename: :obj:`str`
        :returns: (image data, image data type, image shape)
        :rtype: (:class:`numpy.ndarray`, :obj:`str`, :obj:`list` <:obj:`int`>)
        """
        try:
            dtype = None
            shape = None
            nxsfile = filewriter.open_file(
                filename, readonly=True, writer=self.__wrmodule)
            root = nxsfile.root()
            image = root.open("data")
            idata = image.read()
            if image is not None:
                idata = image[...]
                dtype = image.dtype
                shape = image.shape
            nxsfile.close()
            return idata, dtype, shape
        except Exception as e:
            print(str(e))
            if not self.__skipmissing:
                raise Exception("Cannot open a file %s" % filename)
            else:
                print("Cannot open a file %s" % filename)
            return None, None, None

    def _addattr(self, node, attrs):
        """ adds attributes to the parent node in nexus file

        :param node: parent hdf5 node
        :type node: parent hdf5 node
        :param attrs: dictionary with attributes
        """
        attrs = attrs or {}
        for name, (value, dtype, shape) in attrs.items():
            if not self.__testmode:
                node.attributes.create(
                    name, dtype, shape, overwrite=True)[...] = value
            print(" + add attribute: %s = %s" % (name, value))

    def _getfield(self, node, fieldname, dtype, shape, fieldattrs,
                  fieldcompression):
        """ creates a field in nexus file

        :param node: parent hdf5 node
        :type node: :class:`pni.io.nx.h5.nxgroup` or \
                    :class:`pni.io.nx.h5.nxlink`
        :param fieldname: field name
        :type fieldname: :obj:`str`
        :param dtype: field data type
        :type dtype: :obj:`str`
        :param shape: filed data shape
        :type shape: :obj:`list` <:obj:`int`>
        :param fieldattrs: dictionary with field attributes
        :type fieldattrs: :obj:`dict` <:obj:`str`, :obj:`str`>
        :param fieldcompression: field compression rate
        :type fieldcompression: :obj:`int`
        :returns: hdf5 field node
        :rtype: :class:`pni.io.nx.h5.nxfield`
        """
        field = None
        if fieldname in node.names():
            return node.open(fieldname)
        else:
            if not self.__testmode:
                cfilter = None
                if fieldcompression:
                    opts = getcompression(fieldcompression)
                    if isinstance(opts, int):
                        cfilter = filewriter.data_filter(node)
                        cfilter.rate = opts
                    elif isinstance(opts, list) and opts:
                        cfilter = filewriter.data_filter(node)
                        cfilter.filterid = opts[0]
                        cfilter.options = tuple(opts[1:])
                field = node.create_field(
                    fieldname,
                    dtype,
                    shape=[0, shape[0], shape[1]],
                    chunk=[1, shape[0], shape[1]],
                    dfilter=cfilter)
                self._addattr(field, fieldattrs)
            return field

    def _collectimages(self, files, node, fieldname=None, fieldattrs=None,
                       fieldcompression=None, datatype=None, shape=None):
        """ collects images

        :param files: a list of file strings
        :type files: :obj:`list` <:obj:`str`>
        :param node: hdf5 parent node
        :type node: :class:`pni.io.nx.h5.nxgroup` or \
                    :class:`pni.io.nx.h5.nxlink`
        :param fieldname: field name
        :type fieldname: :obj:`str`
        :param fieldattrs: dictionary with field attributes
        :type fieldattrs: :obj:`dict` <:obj:`str`, :obj:`str`>
        :param fieldcompression: field compression rate
        :type fieldcompression: :obj:`int`
        :param datatype: field data type
        :type datatype: :obj:`str`
        :param shape: field shape
        :type shape: :obj:`list` <:obj:`int` >
        """
        fieldname = fieldname or "data"
        field = None
        ind = 0

        for filestr in files:
            if self.__break:
                break
            inputfiles = self._filegenerator(filestr)
            for fname in inputfiles():
                if self.__break:
                    break
                fname = self._findfile(fname, node.name)
                if not fname:
                    continue
                if datatype:
                    data, dtype, shape = self._loadrawimage(
                        fname, datatype, shape)
                elif not fname.endswith(".h5"):
                    data, dtype, shape = self._loadimage(fname)
                else:
                    data, dtype, shape = self._loadh5data(fname)
                if data is not None:
                    if field is None:
                        field = self._getfield(
                            node, fieldname, dtype, shape,
                            fieldattrs, fieldcompression)

                    if field and ind == field.shape[0]:
                        if not self.__testmode:
                            field.grow(0, 1)
                            field[-1, ...] = data
                        print(" * append %s " % (fname))
                    ind += 1
                    if not self.__testmode:
                        self.__nxsfile.flush()

    def _inspect(self, parent, collection=False):
        """ collects recursively the all image files defined
        by hdf5 postrun fields bellow hdf5 parent node

        :param parent: hdf5 parent node
        :type parent: :class:`pni.io.nx.h5.nxgroup` or \
                      :class:`pni.io.nx.h5.nxlink`
        :param collection: if parent is of NXcollection type
        :type collection: :obj:`bool`
        """
        if hasattr(parent, "names"):
            if collection:
                if "postrun" in parent.names():
                    inputfiles = parent.open("postrun")
                    files = inputfiles[...]
                    if isinstance(files, (str, unicode)):
                        files = [files]
                    fieldname = "data"
                    fielddtype = None
                    fieldshape = None
                    fieldattrs = {}
                    fieldcompression = None
                    for at in inputfiles.attributes:
                        if at.name == "fieldname":
                            fieldname = at[...]
                        elif at.name == "fieldcompression":
                            fieldcompression = at[...]
                        elif at.name == "fielddtype":
                            fielddtype = at[...]
                        elif at.name == "fieldshape":
                            fieldshape = json.loads(at[...])
                        elif at.name.startswith("fieldattr_"):
                            atname = at.name[10:]
                            if atname:
                                fieldattrs[atname] = (
                                    at[...], at.dtype, at.shape
                                )

                    print("populate: %s/%s with %s" % (
                        parent.parent.path, fieldname, files))
                    if fieldcompression is None:
                        fieldcompression = self.__compression
                    self._collectimages(
                        files, parent.parent, fieldname, fieldattrs,
                        fieldcompression, fielddtype, fieldshape)
            try:
                names = parent.names()
            except Exception:
                names = []
            for name in names:
                coll = False
                child = parent.open(name)
                if hasattr(child, "attributes"):
                    for at in child.attributes:
                        if at.name == "NX_class":
                            gtype = at[...]
                            if gtype == 'NXcollection':
                                coll = True
                    self._inspect(child, coll)

    def collect(self):
        """ creates a temporary file,
        collects the all image files defined by hdf5
        postrun fields of NXcollection groups and renames the temporary file
        to the origin one if the action was successful
        """
        self._createtmpfile()
        try:
            self.__nxsfile = filewriter.open_file(
                self.__tempfilename, readonly=self.__testmode,
                writer=self.__wrmodule)
            root = self.__nxsfile.root()
            try:
                self.__fullfilename = root.attributes['file_name'][...]
                # print self.__fullfilename
            except Exception:
                pass
            self._inspect(root)
            self.__nxsfile.close()
            if self.__storeold:
                self._storeoldfile()
            shutil.move(self.__tempfilename, self.__nexusfilename)
        except Exception as e:
            print(str(e))
            os.remove(self.__tempfilename)


class Execute(Runner):

    """ Execute runner
    """

    #: (:obj:`str`) command description
    description = "execute the collection process"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxscollect execute -c1 /tmp/gpfs/raw/scan_234.nxs \n\n" \
        + "       nxscollect execute -c32008:0,2 /ramdisk/scan_123.nxs \n\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-c", "--compression", dest="compression",
            action="store", type=str, default=2,
            help="deflate compression rate from 0 to 9 (default: 2)"
            " or <filterid>:opt1,opt2,..."
            " e.g.  -c 32008:0,2  for bitshuffle with lz4")
        parser.add_argument(
            "-s", "--skip_missing", action="store_true",
            default=False, dest="skipmissing",
            help="skip missing files")
        parser.add_argument(
            "-r", "--replace_nexus_file", action="store_true",
            default=False, dest="replaceold",
            help="if it is set the old file is not copied into "
            "a file with .__nxscollect__old__* extension")
        parser.add_argument(
            "--pni", action="store_true",
            default=False, dest="pni",
            help="use pni module as a nexus reader/writer")
        parser.add_argument(
            "--h5py", action="store_true",
            default=False, dest="h5py",
            help="use h5py module as a nexus reader/writer")
        self._parser.add_argument(
            "--h5cpp", action="store_true",
            default=False, dest="h5cpp",
            help="use h5cpp module as a nexus reader")

    def postauto(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument('args', metavar='nexus_file',
                            type=str, nargs='*',
                            help='nexus files to be collected')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        parser = self._parser
        nexusfiles = options.args

        try:
            getcompression(options.compression)
        except Exception as e:
            print(str(e))
            parser.print_help()
            print("")
            sys.exit(0)

        if not nexusfiles or not nexusfiles[0]:
            parser.print_help()
            print("")
            sys.exit(0)

        if options.h5cpp:
            writer = "h5cpp"
        elif options.h5py:
            writer = "h5py"
        elif options.pni:
            writer = "pni"
        elif "h5cpp" in WRITERS.keys():
            writer = "h5cpp"
        elif "h5py" in WRITERS.keys():
            writer = "h5py"
        else:
            writer = "pni"
        if (options.pni and options.h5py and options.h5cpp) or \
           writer not in WRITERS.keys():
            sys.stderr.write("nxscollect: Writer '%s' cannot be opened\n"
                             % writer)
            sys.stderr.flush()
            parser.print_help()
            sys.exit(255)

        # configuration server
        for nxsfile in nexusfiles:
            collector = Collector(
                nxsfile, options.compression, options.skipmissing,
                not options.replaceold, False, writer=writer)
            collector.collect()


class Test(Execute):

    """ Test runner"""

    #: (:obj:`str`) command description
    description = "execute the process in the test mode " \
                  + "without changing any files"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxscollect test /tmp/gpfs/raw/scan_234.nxs \n\n" \
        + "\n"

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        parser = self._parser
        nexusfiles = options.args

        try:
            getcompression(options.compression)
        except Exception as e:
            print(str(e))
            parser.print_help()
            print("")
            sys.exit(0)

        if not nexusfiles or not nexusfiles[0]:
            parser.print_help()
            print("")
            sys.exit(0)

        if options.h5cpp:
            writer = "h5cpp"
        elif options.h5py:
            writer = "h5py"
        elif options.pni:
            writer = "pni"
        elif "h5cpp" in WRITERS.keys():
            writer = "h5cpp"
        elif "h5py" in WRITERS.keys():
            writer = "h5py"
        else:
            writer = "pni"
        if (options.pni and options.h5py and options.h5cpp) or \
           writer not in WRITERS.keys():
            sys.stderr.write("nxscollect: Writer '%s' cannot be opened\n"
                             % writer)
            parser.print_help()
            sys.exit(255)

        # configuration server
        for nxsfile in nexusfiles:
            collector = Collector(
                nxsfile, options.compression, options.skipmissing,
                not options.replaceold, True, writer=writer)
            collector.collect()


def _supportoldcommands():
    """ replace the old command names to the new ones
    """

    oldnew = {
        '-x': 'execute',
        '--execute': 'execute',
        '-t': 'test',
        '--test': 'test',
    }

    if sys.argv and len(sys.argv) > 1:
        if sys.argv[1] in oldnew.keys():
            sys.argv[1] = oldnew[sys.argv[1]]


def main():
    """ the main program function
    """
    description = "  Command-line tool to merge images of external " \
                  + "file-formats into the master NeXus file"

    epilog = 'For more help:\n  nxscollect <sub-command> -h'
    _supportoldcommands()

    parser = NXSArgParser(
        description=description, epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.cmdrunners = [
        ('execute', Execute),
        ('test', Test)
    ]
    runners = parser.createSubParsers()

    try:
        options = parser.parse_args()
    except ErrorException as e:
        sys.stderr.write("Error: %s\n" % str(e))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    if options.subparser is None:
        sys.stderr.write(
            "Error: %s\n" % str("too few arguments"))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    if not WRITERS:
        sys.stderr.write("nxsfileinfo: Neither h5cpp/pni nor h5py installed\n")
        sys.stderr.flush()
        parser.print_help()
        sys.exit(255)

    runners[options.subparser].run(options)


if __name__ == "__main__":
    main()
